#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to visualize Image Dataset.

Methods
-------
    render(items: List[LabeledImageItem], options: Dict = None)
        Renders Labeled Image dataset.

Examples
--------
>>> bbox1 = BoundingBoxItem(bottom_left=(0.3, 0.4),
>>>                        top_left=(0.3, 0.09),
>>>                        top_right=(0.86, 0.09),
>>>                        bottom_right=(0.86, 0.4),
>>>                        labels=['dolphin', 'fish'])

>>> record1 = LabeledImageItem(img_obj1, [bbox1])

>>> bbox2 = BoundingBoxItem(bottom_left=(0.2, 0.4),
>>>                        top_left=(0.2, 0.2),
>>>                        top_right=(0.8, 0.2),
>>>                        bottom_right=(0.8, 0.4),
>>>                        labels=['dolphin'])
>>> bbox3 = BoundingBoxItem(bottom_left=(0.5, 1.0),
>>>                        top_left=(0.5, 0.8),
>>>                        top_right=(0.8, 0.8),
>>>                        bottom_right=(0.8, 1.0),
>>>                        labels=['shark'])

>>> record2 = LabeledImageItem(img_obj2, [bbox2, bbox3])
>>> render(items = [record1, record2], options={"default_color":"blue", "colors": {"dolphin":"blue", "whale":"red"}})
"""

from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Tuple

from matplotlib.colors import is_color_like
import matplotlib.pyplot as plt
from ads.common import logger
import os
import numpy as np
import pandas as pd
from ads.data_labeling.boundingbox import BoundingBoxItem
from PIL.ImageFile import ImageFile
from ads.data_labeling.constants import AnnotationType

DEFAULT_COLOR = "white"
IMG_FORMAT = [".jpg", ".jpeg", ".png"]


class WrongEntityFormat(ValueError):
    def __init__(self):
        super().__init__(
            "Invalid labels from the dataset, "
            f"cannot construct a valid BoundingBoxItem."
        )


@dataclass
class LabeledImageItem:
    """Data class representing Image Item.

    Attributes
    ----------
    img: ImageFile
       the labeled image object.
    boxes: List[BoundingBoxItem]
        a list of BoundingBoxItem
    """

    img: ImageFile
    boxes: List[BoundingBoxItem]

    def _validate(self):
        """Validates the instance.

        Raises
        ------
        ValueError
            If image object is empty.
        WrongEntityFormat
            If the list of entities has a wrong format.
        """
        if self.img is None:
            raise ValueError("The parameter `img` is required.")

        if not isinstance(self.img, ImageFile):
            raise ValueError(
                "The parameter `img` must be an object of type `PIL.ImageFile.ImageFile`."
            )

        if any(not isinstance(entity, BoundingBoxItem) for entity in self.boxes):
            raise WrongEntityFormat()

    def __post_init__(self):
        self._validate()


@dataclass
class RenderOptions:
    """Data class representing render options.

    Attributes
    ----------
    default_color: str
        The specified default color.
    colors: Optional[dict]
        The multiple specified colors.
    """

    default_color: str
    colors: Optional[dict]

    @staticmethod
    def _validate(options: dict) -> None:
        """Validate whether the options passed in fits the defined schema.

        Parameters
        ----------
        options: dict
            The multiple specified colors.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If color provided is not valid.
        """
        if not options:
            return None

        colorvalues = set(options.get("colors", {}).values())
        if "default_color" in options:
            colorvalues.add(options["default_color"])

        for colorval in colorvalues:
            if not is_color_like(colorval):
                raise ValueError(
                    f"{colorval} is not supported. "
                    f"Use RGB format for colors. For instance: `'#EEEEEE'` or `'green'`."
                )

    @classmethod
    def from_dict(cls, options: dict) -> "RenderOptions":
        """Constructs an instance of RenderOptions from a dictionary.

        Parameters
        ----------
        options: dict
            Render options in dictionary format.

        Returns
        -------
        RenderOptions
            The instance of RenderOptions.
        """
        if not options:
            return cls(default_color=DEFAULT_COLOR, colors={})

        RenderOptions._validate(options)

        return cls(
            options.get("default_color", DEFAULT_COLOR), options.get("colors", {}) or {}
        )

    def to_dict(self):
        """Converts RenderOptions instance to dictionary format.

        Returns
        -------
        dict
            The render options in dictionary format.
        """
        return asdict(self)

    def __repr__(self) -> str:
        return repr(self.to_dict())


class ImageLabeledDataFormatter:
    """The ImageRender class to render Image items in a notebook session."""

    @staticmethod
    def render_item(
        item: LabeledImageItem, options: Dict = None, path: str = None
    ) -> None:
        """Renders image dataset.

        Parameters
        ----------
        item: LabeledImageItem
            Item to render.
        options: Optional[dict]
            Render options.
        path: str
            Path to save the image with annotations to local directory.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If items not provided.
            If path is not valid.
        TypeError
            If items provided in a wrong format.
        """
        if not item:
            raise ValueError("The parameter `item` is required.")

        if not isinstance(item, LabeledImageItem):
            raise TypeError(
                "The parameter `item` must be an object of type `LabeledImageItem`."
            )

        render_options = RenderOptions.from_dict(options)

        if path:
            if os.path.isdir(path):
                path += "1.jpg"
            elif not path.lower().endswith(tuple(IMG_FORMAT)):
                img_format_str = (
                    IMG_FORMAT[0]
                    if len(IMG_FORMAT) == 1
                    else ", ".join(IMG_FORMAT[:-1] + ["and " + IMG_FORMAT[-1]])
                )
                raise ValueError(
                    f"Invalid {path}. It is not a directory or the image format "
                    f"in {path} is not supported. Currently the support types "
                    f"are `{img_format_str}`."
                )

        # drow the image with annotations
        ImageLabeledDataFormatter()._draw_labels(
            item=item, options=render_options, path=path
        )

    def _draw_labels(
        self,
        item: LabeledImageItem,
        options: Dict = None,
        path: str = None,
        figure_size: Tuple = (6, 8),
        fontsize: int = 14,
    ):
        """Draw image with annotations.

        Parameters
        ----------
        item: LabeledImageItem
            Item to render.
        options: Optional[dict]
            Render options.
        path: str
            Path to save the image with annotations to local directory.
        figure_size: Tuple
            Figure size of the rendered image.
        fontsize: int
            Font size of the annotations.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            If image type is not PIL.ImageFile.ImageFile.
        """
        img = item.img

        # using matplotlib to open image return numpy array
        if hasattr(img, "shape"):
            im_height, im_width, _ = img.shape
        elif hasattr(img, "size"):
            im_width, im_height = img.size
        else:
            raise TypeError("The input image type must be `PIL.ImageFile.ImageFile`.")

        fig, ax = plt.subplots(1, 1, figsize=figure_size)
        ax.imshow(img)

        for ent in item.boxes:
            category_name = ", ".join([str(label) for label in ent.labels])
            left, top, width, height = self._calculate_bbx(im_width, im_height, ent)

            if len(ent.labels) > 1:
                # sort the multiple labels and use the tuple as key to look up in colormap
                # note: tuple of labels provided as key in colormap must be sorted.
                color_key = tuple(sorted(ent.labels))
                color = options.colors.get(color_key, options.default_color)
            else:
                color = options.colors.get(category_name, options.default_color)

            rect = plt.Rectangle(
                (left, top),
                width,
                height,
                fill=False,
                linewidth=2,
                edgecolor=color,
            )
            ax.add_patch(rect)
            props = dict(boxstyle="round", facecolor=color, alpha=0.6)
            ax.text(
                left,
                top,
                category_name,
                fontsize=fontsize,
                color="black",
                verticalalignment="top",
                bbox=props,
            )

        ax.axis("off")
        if path:
            plt.savefig(path, bbox_inches="tight")
            logger.info(f"The annotated image file is saved in {path}.")

            # Add this line to not show image
            plt.close(fig)
            return

        plt.show()

    def _calculate_bbx(self, im_width, im_height, bbox):
        """calculate bounding box coordinates

        Parameters
        ----------
        im_width: float
            width of the image in pixels
        im_height: float
            height of the image in pixels
        bbox: BoundingBoxItem

        Returns
        -------
        Tuple
            left, top, width, height of the image
        """
        left = bbox.top_left[0] * im_width
        top = bbox.top_left[1] * im_height
        width = (bbox.top_right[0] - bbox.top_left[0]) * im_width
        height = (bbox.bottom_left[1] - bbox.top_left[1]) * im_height
        return left, top, width, height

    def _convert_pil_to_nparray(self, img):
        """convert pil image object to numpy array

        Parameters
        ----------
        img: PIL.ImageFile.ImageFile

        Returns
        -------
        numpy.ndarray
        """
        return np.array(img)


def _df_to_bbox_items(
    df: pd.DataFrame,
    content_column="Content",
    annotations_column: str = "Annotations",
    categories: List[str] = None,
) -> List[LabeledImageItem]:
    """Converts pandas dataframe into a list of LabeledImageItem objects.

    Parameters
    ----------
    df: pd.DataFrame
        The Pandas dataframe to convert.
    content_column: Optional[str]
            The column name with the content data.
    annotations_column: Optional[str]
        The column name for the annotations list.
    categories: Optional List[str]
        The list of object categories in proper order for model training.
        Only used when bounding box annotations are in YOLO format.
            Example: ['cat','dog','horse']

    Returns
    -------
    List[LabeledImageItem]
        The list of LabeledImageItem objects.

    Raises
    ------
    TypeError
        If input data is not a pandas dataframe.
    ValueError
        If input data has a wrong format.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("The parameter `df` must be a Pandas dataframe.")
    if content_column not in list(df.columns):
        raise ValueError(
            "Wrong format of input dataframe. It must have "
            f"`{content_column}` column."
        )
    if annotations_column not in list(df.columns):
        raise ValueError(
            "Wrong format of input dataframe. It must have "
            f"`{annotations_column}` column."
        )

    if df[content_column].isnull().values.any():
        logger.warning(
            "The source Dataframe includes records where content is not loaded. "
            "Use `materialize=True` to load the content. "
            "The records with empty content will be ignored."
        )

    result = []
    for item in df.T.to_dict().values():
        if item[annotations_column] and not isinstance(item[annotations_column], list):
            raise ValueError(
                "The parameter `df` is invalid. "
                f"The column {annotations_column}  must be of type `List[BoundingBoxItem]`."
            )
        if item[content_column]:
            if (
                isinstance(item[annotations_column][0], list)
                and item[annotations_column][0][0]
                and isinstance(item[annotations_column][0][0], Tuple)
                and len(item[annotations_column][0][0]) == 5
            ):
                bbox_items = [
                    BoundingBoxItem.from_yolo(bbox, categories)
                    for bbox in item[annotations_column]
                ]
            else:
                bbox_items = item[annotations_column] or []
            result.append(LabeledImageItem(item[content_column], bbox_items))

    return result


def render(
    items: List[LabeledImageItem], options: Dict = None, path: str = None
) -> None:
    """Render image dataset.

    Parameters
    ----------
    items: List[LabeledImageItem]
        The list of LabeledImageItem to render.
    options: dict, optional
        The options for rendering.
    path: str
        Path to save the images with annotations to local directory.

    Returns
    -------
    None
        Nothing.

    Raises
    ------
    ValueError
        If items not provided.
        If path is not valid.
    TypeError
        If items provided in a wrong format.

    Examples
    --------
    >>> bbox1 = BoundingBoxItem(bottom_left=(0.3, 0.4),
    >>>                        top_left=(0.3, 0.09),
    >>>                        top_right=(0.86, 0.09),
    >>>                        bottom_right=(0.86, 0.4),
    >>>                        labels=['dolphin', 'fish'])

    >>> record1 = LabeledImageItem(img_obj1, [bbox1])
    >>> render(items = [record1])
    """
    if not items:
        raise ValueError("The parameter `items` is required.")

    if not isinstance(items, list) or not all(
        isinstance(x, LabeledImageItem) for x in items
    ):
        raise TypeError(
            "Wrong format for the items. The items must be `List[LabeledImageItem]`."
        )

    for idx, item in enumerate(items):
        if not path:
            ImageLabeledDataFormatter.render_item(item, options)
        else:
            if os.path.isdir(path):
                ImageLabeledDataFormatter.render_item(
                    item, options, path=f"{path}_{idx+1}.jpg"
                )
            elif not path.lower().endswith(tuple(IMG_FORMAT)):
                img_format_str = (
                    IMG_FORMAT[0]
                    if len(IMG_FORMAT) == 1
                    else ", ".join(IMG_FORMAT[:-1] + ["and " + IMG_FORMAT[-1]])
                )
                raise ValueError(
                    f"Invalid {path}. It is not a directory or the image format "
                    f"in {path} is not supported. Currently the support "
                    f"types are {img_format_str}."
                )
            else:
                img_type = path.split(".")[-1]
                path_root = ".".join(path.split(".")[:-1])
                new_path = f"{path_root}_{idx+1}.{img_type}"
                ImageLabeledDataFormatter.render_item(item, options, path=new_path)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class ClusterProviderFactory:
    """
    Factory class for creating provider instance.
    """

    provider = {}

    @classmethod
    def register(cls, cluster_type, provider_class):
        ClusterProviderFactory.provider[cluster_type.upper()] = provider_class

    @staticmethod
    def get_provider(key, *args, **kwargs):
        return ClusterProviderFactory.provider[key](*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from dataclasses import dataclass, field
from typing import List, Tuple


@dataclass
class BoundingBoxItem:
    """BoundingBoxItem class representing bounding box label.

    Attributes
    ----------
    labels: List[str]
        List of labels for this bounding box.
    top_left: Tuple[float, float]
        Top left corner of this bounding box.
    bottom_left: Tuple[float, float]
        Bottom left corner of this bounding box.
    bottom_right: Tuple[float, float]
        Bottom right corner of this bounding box.
    top_right: Tuple[float, float]
        Top right corner of this bounding box.

    Examples
    --------
    >>> item = BoundingBoxItem(
    ...     labels = ['cat','dog']
    ...     bottom_left=(0.2, 0.4),
    ...     top_left=(0.2, 0.2),
    ...     top_right=(0.8, 0.2),
    ...     bottom_right=(0.8, 0.4))
    >>> item.to_yolo(categories = ['cat','dog', 'horse'])
    """

    top_left: Tuple[float, float]
    bottom_left: Tuple[float, float]
    bottom_right: Tuple[float, float]
    top_right: Tuple[float, float]
    labels: List[str] = field(default_factory=list)

    def _validate(self):
        """Validates the instance.

        Raises
        ------
        ValueError
            If the bounding box coordinate is not between [0.0, 1.0].
        """
        if (
            not self.labels
            or not isinstance(self.labels, list)
            or len(self.labels) == 0
        ):
            raise ValueError(
                "The parameter `df` is invalid. The BoundingBoxItem objects in the column "
                f"{self.labels}  must contain a non-empty list of string labels."
            )

        if any(
            (
                not isinstance(entity, Tuple)
                or len(entity) != 2
                or not isinstance(entity[0], float)
                or not isinstance(entity[1], float)
                or not 0.0 <= entity[0] <= 1.0
                or not 0.0 <= entity[1] <= 1.0
            )
            for entity in [
                self.bottom_left,
                self.top_left,
                self.top_right,
                self.bottom_right,
            ]
        ):
            raise ValueError(
                "The parameter `df` is invalid. The BoundingBoxItem objects in the column "
                f"{[self.bottom_left, self.top_left, self.top_right, self.bottom_right,]} must "
                "contain a tuple of two real numbers in the range of [0, 1]. "
                "One BoundingBoxItem contains the values (invalid_tuple). Use the `DataLabeling.export()` "
                "method to create a new dataset record file."
            )

    def __post_init__(self):
        self._validate()

    def to_yolo(
        self, categories: List[str]
    ) -> List[Tuple[int, float, float, float, float]]:
        """Converts BoundingBoxItem to the YOLO format.

        Parameters
        ----------
        categories: List[str]
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        List[Tuple[int, float, float, float, float]]
            The list of YOLO formatted bounding boxes.

        Raises
        ------
        ValueError
            When categories list not provided.
            When categories list not matched with the labels.
        TypeError
            When categories list has a wrong format.
        """
        if not categories:
            raise ValueError(
                "The parameter `categories` is required. Use the `.info()` method to obtain a list of categories for this dataset."
            )
        if not isinstance(categories, list):
            raise TypeError(
                "The parameter `categories` is invalid. It must be an object of type `List[str]`."
            )
        if not set(self.labels).issubset(categories):
            raise ValueError(
                "The parameter `categories` is invalid. It must be a list of all the unique labels in the dataset."
            )

        category_map = {k: v for v, k in enumerate(categories)}
        coords = (
            (self.top_left[0] + self.top_right[0]) / 2,
            (self.top_left[1] + self.bottom_left[1]) / 2,
            self.top_right[0] - self.top_left[0],
            self.bottom_left[1] - self.top_left[1],
        )

        return [(category_map[label],) + coords for label in self.labels]

    @classmethod
    def from_yolo(
        cls, bbox: List[Tuple], categories: List[str] = None
    ) -> "BoundingBoxItem":
        """Converts the YOLO formated annotations to BoundingBoxItem.

        Parameters
        ----------
        bboxes: List[Tuple]
            The list of bounding box annotations in YOLO format.
            Example: [(0, 0.511560675, 0.50234826, 0.47013485, 0.57803468)]
        categories: List[str]
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        BoundingBoxItem
            The BoundingBoxItem.

        Raises
        ------
        TypeError
            When categories list has a wrong format.
        """

        if bbox:
            _, x1, y1, w_size, h_size = bbox[0]
            labels = [label[0] for label in bbox]
            if categories:
                if not isinstance(categories, list):
                    raise TypeError("The categories must be a List[str].")
                if max(labels) + 1 > len(categories):
                    raise ValueError(
                        "Index out of the range of the categories list. "
                        "The categories must contain all the labels "
                        "in the order that the integer label corresponds to."
                    )

                labels = [categories[label[0]] for label in bbox]

            top_left = (x1 - (w_size / 2), y1 - (h_size / 2))
            top_right = (x1 + (w_size / 2), y1 - (h_size / 2))
            bottom_right = (x1 + (w_size / 2), y1 + (h_size / 2))
            bottom_left = (x1 - (w_size / 2), y1 + (h_size / 2))
            bbox_item = BoundingBoxItem(
                top_left, bottom_left, bottom_right, top_right, labels
            )
            bbox_item._validate()
            return bbox_item
        else:
            return None


@dataclass
class BoundingBoxItems:
    """BoundingBoxItems class which consists of a list of BoundingBoxItem.

    Attributes
    ----------
    items: List[BoundingBoxItem]
        List of BoundingBoxItem.

    Examples
    --------
    >>> item = BoundingBoxItem(
    ...     labels = ['cat','dog']
    ...     bottom_left=(0.2, 0.4),
    ...     top_left=(0.2, 0.2),
    ...     top_right=(0.8, 0.2),
    ...     bottom_right=(0.8, 0.4))
    >>> items = BoundingBoxItems(items = [item])
    >>> items.to_yolo(categories = ['cat','dog', 'horse'])
    """

    items: List[BoundingBoxItem] = field(default_factory=list)

    def __getitem__(self, index: int) -> BoundingBoxItem:
        return self.items[index]

    def to_yolo(
        self, categories: List[str]
    ) -> List[Tuple[int, float, float, float, float]]:
        """Converts BoundingBoxItems to the YOLO format.

        Parameters
        ----------
        categories: List[str]
            The list of object categories in proper order for model training.
            Example: ['cat','dog','horse']

        Returns
        -------
        List[Tuple[int, float, float, float, float]]
            The list of YOLO formatted bounding boxes.

        Raises
        ------
        ValueError
            When categories list not provided.
            When categories list not matched with the labels.
        TypeError
            When categories list has a wrong format.
        """
        if not categories:
            raise ValueError(
                "The parameter `categories` is required. Use the `.info()` method to obtain a list of categories for this dataset."
            )
        if not isinstance(categories, list):
            raise TypeError(
                "The parameter `categories` is invalid. It must be an object of type `List[str]`."
            )

        result = []
        for item in self.items:
            result.extend(item.to_yolo(categories))
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents an IpAddressV4 feature type.

Classes:
    IpAddressV4
        The IpAddressV4 feature type.
"""
import re

import pandas as pd
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import _count_unique_missing
from ads.feature_engineering import schema

PATTERN = re.compile(
    r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)",
    re.IGNORECASE,
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """
    return data.apply(
        lambda x: True
        if not pd.isnull(x) and PATTERN.match(str(x)) is not None
        else False
    )


class IpAddressV4(FeatureType):
    """
    Type representing IP Address V4.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.

    Example
    -------
    >>> from ads.feature_engineering.feature_type.ip_address_v4 import IpAddressV4
    >>> import pandas as pd
    >>> import numpy as np
    >>> s = pd.Series(['192.168.0.1', '2001:db8::', '', np.NaN, None], name='ip_address')
    >>> s.ads.feature_type = ['ip_address_v4']
    >>> IpAddressV4.validator.is_ip_address_v4(s)
    0     True
    1    False
    2    False
    3    False
    4    False
    Name: ip_address, dtype: bool
    """

    description = "Type representing IP Address V4."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count).

        Examples
        --------
        >>> s = pd.Series(['192.168.0.1', '192.168.0.2', '192.168.0.3', '192.168.0.4', np.NaN, None], name='ip_address')
        >>> s.ads.feature_type = ['ip_address_v4']
        >>> s.ads.feature_stat()
            Metric  Value
        0	count	6
        1	unique	4
        2	missing	2

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series provided.
        """
        return _count_unique_missing(x)

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> s = pd.Series(['192.168.0.1', '192.168.0.2', '192.168.0.3', '192.168.0.4', np.NaN, None], name='ip_address_v4')
        >>> s.ads.feature_type = ['ip_address_v4']
        >>> s.ads.feature_domain()
        constraints: []
        stats:
            count: 6
            missing: 2
            unique: 4
        values: IpAddressV4

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the IpAddressV4 feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


IpAddressV4.validator.register("is_ip_address_v4", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.data_labeling.data_labeling_service import DataLabeling
from ads.data_labeling.loader.file_loader import ImageFileLoader, TextFileLoader
from ads.data_labeling.metadata import Metadata
from ads.data_labeling.reader.dataset_reader import LabeledDatasetReader
from ads.data_labeling.reader.metadata_reader import MetadataReader
from ads.data_labeling.reader.record_reader import RecordReader
from ads.data_labeling.constants import DatasetType, AnnotationType



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from ads.explanations import MLXGlobalExplainer, MLXLocalExplainer
from ads.explanations.mlx_whatif_explainer import MLXWhatIfExplainer
from ads.explanations.mlx_interface import _reset_index
from ads.common.decorator.deprecate import deprecated


class ADSExplainer(object):
    """
    Main ADS Explainer class for configuring and managing local and global
    explanation objects.
    """

    @deprecated(
        details="Working with AutoML has moved from within ADS to working directly with the AutoMLx library. AutoMLx are preinstalled in conda pack automlx_p38_cpu_v2 and later, and can now be updated independently of ADS. AutoMLx documentation may be found at https://docs.oracle.com/en-us/iaas/tools/automlx/latest/html/multiversion/v23.1.1/index.html. Notebook examples are in Oracle's samples repository: https://github.com/oracle-samples/oci-data-science-ai-samples/tree/master/notebook_examples and a migration tutorial can be found at https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_training/automl/quick_start.html .",
        raise_error=True,
    )
    def __init__(self, test_data, model, training_data=None):
        """
        Creates an ads explainer object.

        Parameters
        ----------
        test_data : ads.common.data.ADSData
            The test data for exaplanation built with `ADSData.build()`.
        model : ads.common.model.ADSModel
            The model used for explanation built with `ADSModel.from_estimator()`.
            Maximum length of the list is 3.
        training_data : ads.common.data.ADSData instance, optional
            Training data using `ADSData.build()`.
            Using `ADSData.build()`.

        Examples
        --------
        train, test = ds.train_test_split()

        model = MyModelClass.train(train)

        explainer = ADSExplainer(test, model)
        """

        self.X_train = None
        self.y_train = None
        if training_data is not None:
            self.X_train, self.y_train = training_data.X, training_data.y
        self.X_test, self.y_test = test_data.X, test_data.y
        self.model = model

    def global_explanation(self, provider=MLXGlobalExplainer()):
        """
        Initializes the given global explainer provider and returns it. The returned
        explainer will have the following properties set:

        * **est**: Estimator in the ADSModel
        * **class_names** (`Iterable`): Target classes. None for regression
        * **X_test** (`pandas.DataFrame`): Test data to explain.
        * **y_test** (`pandas.Series`): Labels for test data.
        * **X_train** (`pandas.DataFrame`, optional): Training data to build explainer.
        * **y_train** (`pandas.Series`, optional) Labels for training data.

        Parameters
        ----------
        provider : instance of `GlobalExplainer`, optional
            The explainer instance that is set up. Defaults to `MLXGlobalExplainer`.
            The properties will be set on the object

        Return
        ------
        `GlobalExplainer` object
            Modified instance of `GlobalExplainer` implementation passed into function

        Examples
        --------
        explainer = ADSExplainer(test, model)

        global_explanation = explainer.global_explanation()
        """

        provider.setup(self.model, self.X_test, self.y_test, self.X_train, self.y_train)
        return provider

    def local_explanation(self, provider=MLXLocalExplainer()):
        """Initializes the given local explainer provider and returns it.

        The returned `LocalExplainer` instance has the following properties set:

        * **est**: Estimator in the ADSModel.
        * **class_names** (`Iterable`): Target classes. None for regression.
        * **X_test** (`pandas.DataFrame`): Test data to explain.
        * **y_test** (`pandas.Series`): Labels for test data.
        * **X_train** (`pandas.DataFrame`, optional): Training data to build explainer.
        * **y_train** (`pandas.Series`, optional): Labels for training data.

        Parameters
        ----------
        provider : instance of `LocalExplainer`, optional
            The explainer instance that is set up. Defaults to `MLXLocalExplainer`.

        Return
        ------
        `LocalExplainer` object
            Modified instance of `LocalExplainer` implementation passed into function.

        Examples
        --------
        explainer = ADSExplainer(test, model)

        local_explanation = explainer.local_explanation()
        """

        provider.setup(self.model, self.X_test, self.y_test, self.X_train, self.y_train)
        return provider

    def whatif_explanation(self, provider=MLXWhatIfExplainer()):
        """Initializes the given mlx explorer provider and returns it.

        The returned `MLXWhatIfExplainer` instance would have the following properties set:

        * **est**: Estimator in the ADSModel.
        * **class_names** (Iterable): Target classes. None for regression.
        * **X_test** (pandas.DataFrame): Test data to explain.
        * **y_test** (pandas.Series): Labels for test data.
        * **X_train** (pandas.DataFrame, optional): Training data to build explainer.
        * **y_train** (pandas.Series, optional): Labels for training data.

        Parameters
        ----------
        provider : Instance of `MLXWhatIfExplainer` implementation, optional.
            Defaults to `MLXWhatIfExplainer`

        Return
        ------
        `MLXWhatIfExplainer` object
            Configured explanation provider.

        Examples
        --------
        explainer = ADSExplainer(test, model)

        local_explanation = explainer.explore_sample()
        """

        if self.X_train is None:
            provider.setup(
                model=self.model,
                X_test=_reset_index(self.X_test),
                y_test=_reset_index(self.y_test),
                X_train=self.X_train,
                y_train=self.y_train,
            )
        else:
            provider.setup(
                model=self.model,
                X_test=_reset_index(self.X_test),
                y_test=_reset_index(self.y_test),
                X_train=_reset_index(self.X_train),
                y_train=_reset_index(self.y_train),
            )
        return provider

    def show_in_notebook(self, provider=MLXGlobalExplainer()):
        """Generates a global or local explanation based on the default values and
        returns a visualization for the explanation.

        Parameters
        ----------
        provider : Instance of `GlobalExplainer` or `LocalExplainer` implementation, optional
            Configures the type of explanation to generate and visualize (local or global).
            Default is `MLXGlobalExplainer()`. Currently does not support `MLXWhatIfExplainer`.

        Return
        ------
        `None`
            Nothing


        Raises
        ------
        TypeError
            If `provider` is an instance of `MLXWhatIfExplainer`
        """

        if isinstance(provider, MLXWhatIfExplainer):
            raise TypeError("MLXWhatIfExplainer does not support show_in_notebook.")
        self.global_explanation(provider=provider).show_in_notebook()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class ChangesNotCommitted(Exception):
    def __init__(self, path):
        msg = f"""
            File(s) at {path} are either dirty or untracked.
            Please commit changes and then save the model, or set `ignore_pending_changes=True`.
        """
        super().__init__(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a DateTime feature type.

Classes:
    DateTime
        The DateTime feature type.
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandas.api.types as pdtypes
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _add_missing,
    _set_seaborn_theme,
    SchemeTeal,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """
    Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """

    def _is_datetime(x: any):
        if pd.isnull(x):
            return False
        if pdtypes.is_datetime64_any_dtype(type(x)):
            return True
        if pdtypes.is_string_dtype(type(x)) or pdtypes.is_object_dtype(type(x)):
            try:
                pd.to_datetime(x)
            except:
                return False
            return True
        return False

    return data.apply(lambda x: True if _is_datetime(x) else False)


class DateTime(FeatureType):
    """
    Type representing date and/or time.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows distributions of datetime datasets using histograms.

    Example
    -------
    >>> from ads.feature_engineering.feature_type.datetime import DateTime
    >>> import pandas as pd
    >>> s = pd.Series(["12/12/12", "12/12/13", None, "12/12/14"], name='datetime')
    >>> s.ads.feature_type = ['date_time']
    >>> DateTime.validator.is_datetime(s)
    0     True
    1     True
    2    False
    3     True
    Name: datetime, dtype: bool
    """

    description = "Type representing date and/or time."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, sample maximum, sample minimum, and
        missing(count) if there is any.

        Examples
        --------
        >>> x = pd.Series(['3/11/2000', '3/12/2000', '3/13/2000', '', None, np.nan, 'April/13/2011', 'April/15/11'], name='datetime')
        >>> x.ads.feature_type = ['date_time']
        >>> x.ads.feature_stat()
            Metric              Value
        0	count	            8
        1	sample maximum	    April/15/11
        2	sample minimum	    3/11/2000
        3	missing	            3

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.
        """
        df_stat = pd.Series(
            {
                "count": len(x),
                "sample maximum": x.replace(r"", np.NaN).dropna().max(),
                "sample minimum": x.replace(r"", np.NaN).dropna().min(),
            },
            name=x.name,
        ).to_frame()
        return _add_missing(x.replace(r"", np.NaN), df_stat)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows distributions of datetime datasets using histograms.

        Examples
        --------
        >>> x = pd.Series(['3/11/2000', '3/12/2000', '3/13/2000', '', None, np.nan, 'April/13/2011', 'April/15/11'], name='datetime')
        >>> x.ads.feature_type = ['date_time']
        >>> x.ads.feature_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the DateTime feature type.
        """
        col_name = x.name if x.name else "datetime"
        df = x.to_frame(col_name)
        df["validation"] = default_handler(x)
        df = df[df["validation"] == True]
        if len(df.index):
            df[col_name] = df[col_name].apply(lambda x: pd.to_datetime(x))
            _set_seaborn_theme()
            return seaborn.histplot(data=df, y=col_name, color=SchemeTeal.AREA_DARK)

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> s = pd.Series(['3/11/2000', '3/12/2000', '3/13/2000', '', None, np.nan, 'April/13/2011', 'April/15/11'], name='datetime')
        >>> s.ads.feature_type = ['date_time']
        >>> s.ads.feature_domain()
        constraints: []
        stats:
            count: 8
            missing: 3
            sample maximum: April/15/11
            sample minimum: 3/11/2000
        values: DateTime

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the DateTime feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [],
        )


DateTime.validator.register("is_datetime", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, division, absolute_import
import os
import logging
import json
from typing import Callable, Dict, Optional, Union

__version__ = ""
with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "ads_version.json")
) as version_file:
    __version__ = json.load(version_file)["version"]
import oci

import matplotlib.font_manager  # causes matplotlib to regenerate its fonts

import ocifs
from ads.common.decorator.deprecate import deprecated
from ads.common.ipython import configure_plotting, _log_traceback
from ads.feature_engineering.accessor.series_accessor import ADSSeriesAccessor
from ads.feature_engineering.accessor.dataframe_accessor import ADSDataFrameAccessor
from ads.common import auth
from ads.common.auth import set_auth

os.environ["GIT_PYTHON_REFRESH"] = "quiet"


debug_mode = os.environ.get("DEBUG_MODE", False)
documentation_mode = os.environ.get("DOCUMENTATION_MODE", "False") == "True"
test_mode = os.environ.get("TEST_MODE", False)
resource_principal_mode = bool(
    os.environ.get("RESOURCE_PRINCIPAL_MODE", False)
)  # deprecated with is_resource_principal_mode() from ads.common.utils
orig_ipython_traceback = None


def getLogger(name="ads"):
    logger = logging.getLogger(name)
    logger.setLevel(logging.WARNING)
    return logger


logger = getLogger(__name__)
logger.addHandler(logging.NullHandler())

# All warnings are logged by default
logging.captureWarnings(True)


def set_debug_mode(mode=True):
    """
    Enable/disable printing stack traces on notebook.

    Parameters
    ----------
    mode: bool (default True)
         Enable/disable print stack traces on notebook

    """
    global debug_mode
    debug_mode = mode
    import IPython

    if debug_mode:
        from ads.common.ipython import orig_ipython_traceback

        IPython.core.interactiveshell.InteractiveShell.showtraceback = (
            orig_ipython_traceback
        )
    else:
        IPython.core.interactiveshell.InteractiveShell.showtraceback = _log_traceback


@deprecated("2.3.1")
def set_documentation_mode(mode=False):
    """
    This method is deprecated and will be removed in future releases.
    Enable/disable printing user tips on notebook.

    Parameters
    ----------
    mode: bool (default False)
        Enable/disable print user tips on notebook
    """
    global documentation_mode
    documentation_mode = mode


@deprecated("2.3.1")
def set_expert_mode():
    """
    This method is deprecated and will be removed in future releases.
    Enables the debug and documentation mode for expert users all in one method.
    """
    set_debug_mode(True)
    set_documentation_mode(False)


#
# ***FOR TESTING PURPOSE ONLY***
#
def _set_test_mode(mode=False):
    """
    Enable/disable intercept the automl call and rewrite it to always use
    the two algorithms (LogisticRegression for classification and LinearRegression for regression).
    Enable only during tests to reduce nb convert notebook tests run time.

    Parameters
    ----------
    mode: bool (default False)
         Enable/disable the ability to intercept automl call

    """
    global test_mode
    test_mode = mode


def hello():
    """
    Imports Pandas, sets the documentation mode, and prints a fancy "Hello".
    """
    import pandas

    global documentation_mode
    global debug_mode

    print(
        f"""

  O  o-o   o-o
 / \\ |  \\ |
o---o|   O o-o
|   ||  /     |
o   oo-o  o--o

ADS SDK version: {__version__}
Pandas version: {pandas.__version__}
Debug mode: {debug_mode}
"""
    )


configure_plotting()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Boolean feature type.

Classes:
    Boolean
        The feature type that represents binary values True/False.

Functions:
    default_handler(data: pd.Series) -> pd.Series
        Processes given data and indicates if the data matches requirements.
"""
import matplotlib.pyplot as plt
import pandas as pd
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _count_unique_missing,
    is_boolean,
    _set_seaborn_theme,
    SchemeTeal,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """
    return pd.Series((is_boolean(value) for value in data.values))


class Boolean(FeatureType):
    """
    Type representing binary values True/False.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Show the counts of observations in True/False using bars.

    Examples
    --------
    >>> from ads.feature_engineering.feature_type.boolean import Boolean
    >>> import pandas as pd
    >>> import numpy as np
    >>> s = pd.Series([True, False, True, False, np.NaN, None], name='bool')
    >>> s.ads.feature_type = ['boolean']
    >>> Boolean.validator.is_boolean(s)
    0     True
    1     True
    2     True
    3     True
    4    False
    5    False
    dtype: bool
    """

    description = "Type representing binary values True/False."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and missing(count).

        Parameters
        ----------
        x : :class:`pandas.Series`
            The feature being evaluated.

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.

        Examples
        --------
        >>> s = pd.Series([True, False, True, False, np.NaN, None], name='bool')
        >>> s.ads.feature_type = ['boolean']
        >>> s.ads.feature_stat()
            Metric  Value
        0	count	6
        1	unique	2
        2	missing	2
        """
        return _count_unique_missing(x)

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows the counts of observations in True/False using bars.

        Parameters
        ----------
        x : :class:`pandas.Series`
            The feature being evaluated.

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Boolean feature type.

        Examples
        --------
        >>> s = pd.Series([True, False, True, False, np.NaN, None], name='bool')
        >>> s.ads.feature_type = ['boolean']
        >>> s.ads.feature_plot()
        """
        col_name = x.name if x.name else "boolean"
        df = x.to_frame(col_name)
        df["validation"] = default_handler(x)
        df = df[df["validation"] == True]
        if len(df.index):
            _set_seaborn_theme()
            ax = seaborn.countplot(y=col_name, data=df, color=SchemeTeal.AREA_DARK)
            ax.set(xlabel="Count")
            return ax

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> s = pd.Series([True, False, True, False, np.NaN, None], name='bool')
        >>> s.ads.feature_type = ['boolean']
        >>> s.ads.feature_domain()
        constraints:
        - expression: $x in [True, False]
            language: python
        stats:
            count: 6
            missing: 2
            unique: 2
        values: Boolean

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the Boolean feature type.
        """

        return schema.Domain(
            cls.__name__,
            cls.feature_stat(x).to_dict()[x.name],
            [schema.Expression("$x in [True, False]")],
        )


Boolean.validator.register("is_boolean", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import re
import time
from urllib.parse import urlparse

import pandas as pd
from ads.common import auth, oci_client, utils
from ads.common.oci_mixin import OCIWorkRequestMixin
from ads.common.utils import snake_to_camel
from ads.config import JOB_RUN_COMPARTMENT_OCID, NB_SESSION_COMPARTMENT_OCID
from ads.common.object_storage_details import (
    ObjectStorageDetails,
    InvalidObjectStoragePath,
)
from oci import pagination
from oci.data_labeling_service.models import (
    ObjectStorageSnapshotExportDetails,
    SnapshotDatasetDetails,
    WorkRequest,
)
from oci.util import to_dict

logger = logging.getLogger(__name__)
NUM_PROGRESS_BAR = 8


class DataLabeling(OCIWorkRequestMixin):
    """Class for data labeling service. Integrate the data labeling service APIs.

    Examples
    --------
    >>> import ads
    >>> import pandas
    >>> from ads.data_labeling.data_labeling_service import DataLabeling
    >>> ads.set_auth("api_key")
    >>> dls = DataLabeling()
    >>> dls.list_dataset()
    >>> metadata_path = dls.export(dataset_id="your dataset id",
    ...     path="oci://<bucket_name>@<namespace>/folder")
    >>> df = pd.DataFrame.ads.read_labeled_data(metadata_path)
    """

    def __init__(
        self,
        compartment_id: str = None,
        dls_cp_client_auth: dict = None,
        dls_dp_client_auth: dict = None,
    ) -> None:
        """Initialize a DataLabeling class.

        Parameters
        ----------
        compartment_id : str, optional
            OCID of data labeling datasets' compartment
        dls_cp_client_auth : dict, optional
            Data Labeling control plane client auth. Default is None. The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        dls_dp_client_auth : dict, optional
            Data Labeling data plane client auth. Default is None. The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        None
            Nothing.
        """
        self.compartment_id = compartment_id
        if self.compartment_id is None:
            self.compartment_id = (
                NB_SESSION_COMPARTMENT_OCID or JOB_RUN_COMPARTMENT_OCID
            )

        if not self.compartment_id:
            raise ValueError("The parameter `compartment_id` is required.")

        self.dls_cp_client_auth = dls_cp_client_auth or auth.default_signer()
        self.dls_dp_client_auth = dls_dp_client_auth or auth.default_signer()

        self.dls_dp_client = oci_client.OCIClientFactory(
            **self.dls_dp_client_auth
        ).data_labeling_dp

        self.dls_cp_client = oci_client.OCIClientFactory(
            **self.dls_cp_client_auth
        ).data_labeling_cp

    def list_dataset(self, **kwargs) -> pd.DataFrame:
        """List all the datasets created from the data labeling service under a given compartment.

        Parameters
        ----------
        kwargs : dict, optional
            Additional keyword arguments will be passed to oci.data_labeling_serviceDataLabelingManagementClient.list_datasets method.

        Returns
        -------
        pandas.DataFrame
            pandas dataframe which contains the dataset information.

        Raises
        ------
        Exception
            If pagination.list_call_get_all_results() fails
        """
        try:
            items = pagination.list_call_get_all_results(
                self.dls_cp_client.list_datasets, self.compartment_id, **kwargs
            ).data
        except Exception as e:
            raise e

        df = pd.DataFrame()
        if items:
            df = pd.concat(
                [
                    pd.DataFrame(to_dict(dataset), index=[i])
                    for i, dataset in enumerate(items)
                ],
                axis=0,
            )
            df = (
                df.reset_index(drop=True).set_index("id").drop(columns="compartment_id")
            )
        df.columns = [
            snake_to_camel(name, capitalized_first_token=True) for name in df.columns
        ]
        return df

    def export(self, dataset_id: str, path: str, include_unlabeled=False) -> str:
        """Export dataset based on the dataset_id and save the jsonl files under the path
        (metadata jsonl file and the records jsonl file) to the object storage path provided by the user
        and return the metadata jsonl path.

        Parameters
        ----------
        dataset_id : str
            The dataset id of which the snapshot will be generated.
        path : str
            The object storage path to store the generated snapshot.
            "oci://<bucket_name>@<namespace>/prefix"
        include_unlabeled: bool, Optional. Defaults to False.
            Whether to include unlabeled records or not.

        Returns
        -------
        str
            oci path of the metadata jsonl file.
        """

        if not re.match(r"oci://*@*", path):
            raise InvalidObjectStoragePath(
                "The parameter `path` is not valid. It must follow the pattern 'oci://<bucket_name>@<namespace>/key'."
            )
        url_parse = urlparse(path)
        bucketname = url_parse.username
        namespace = url_parse.hostname
        if not bucketname:
            raise InvalidObjectStoragePath(
                f"The parameter `path` is not valid. The bucket name ({bucketname}) was not found. It must follow the pattern 'oci://<bucket_name>@<namespace>/key'."
            )
        if not namespace:
            raise InvalidObjectStoragePath(
                f"The parameter `path` is not valid. The name space ({namespace}) was not found. It must follow the pattern 'oci://<bucket_name>@<namespace>/key'."
            )

        prefix = url_parse.path.strip("/")
        self.client = self.dls_cp_client

        if not prefix.endswith("/"):
            prefix = prefix + "/"
        os_snapshot_export_detail = ObjectStorageSnapshotExportDetails(
            export_type="OBJECT_STORAGE",
            namespace=namespace,
            bucket=bucketname,
            prefix=prefix,
        )

        snapshot_detail = SnapshotDatasetDetails(
            are_annotations_included=True,
            are_unannotated_records_included=include_unlabeled,
            export_details=os_snapshot_export_detail,
        )
        try:
            snapshot_detail_response = self.dls_cp_client.snapshot_dataset(
                dataset_id, snapshot_detail
            )
        except Exception as error:
            if dataset_id not in self.list_dataset().index:
                raise ValueError(
                    "The parameter `dataset_id` is invalid. "
                    "Use the `.list_dataset()` method to obtain a list of  all available datasets."
                )
            raise error

        res_work_request = self._wait_for_work_request(snapshot_detail_response)

        metadata = res_work_request.data.resources[1].metadata
        return ObjectStorageDetails(
            metadata["BUCKET"], metadata["NAMESPACE"], metadata["OBJECT"]
        ).path

    def _wait_for_work_request(self, snapshot_detail_response):
        successful_state = WorkRequest.STATUS_SUCCEEDED
        wait_for_states = (
            WorkRequest.STATUS_CANCELED,
            WorkRequest.STATUS_CANCELING,
            WorkRequest.STATUS_FAILED,
        )
        work_request_id = snapshot_detail_response.headers["opc-work-request-id"]
        work_request_log_entires = self.dls_cp_client.list_work_request_logs(
            work_request_id
        ).data.items
        i = 0
        res_work_request = self.dls_cp_client.get_work_request(work_request_id)
        self._num_progress_bar = NUM_PROGRESS_BAR
        with utils.get_progress_bar(self._num_progress_bar) as progress:
            while len(work_request_log_entires) <= self._num_progress_bar - 1:
                new_work_request_log_entires = work_request_log_entires[i:]
                for work_request_log_entry in new_work_request_log_entires:
                    progress.update(work_request_log_entry.message)
                    i += 1
                if (
                    len(work_request_log_entires) == self._num_progress_bar - 1
                    or res_work_request.data.status in wait_for_states
                ):
                    progress.update(work_request_log_entires[-1].message)
                    if res_work_request.data.status != successful_state:
                        raise Exception(work_request_log_entires[-1].message)
                    else:
                        break
                time.sleep(i)
                work_request_log_entires = self.dls_cp_client.list_work_request_logs(
                    work_request_id
                ).data.items
                res_work_request = self.dls_cp_client.get_work_request(work_request_id)
        return res_work_request



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import copy
import os
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

import ads.telemetry
import oci
from ads.common import logger
from ads.common.decorator.deprecate import deprecated
from ads.common.extended_enum import ExtendedEnumMeta
from oci.config import DEFAULT_LOCATION  # "~/.oci/config"
from oci.config import DEFAULT_PROFILE  # "DEFAULT"


class AuthType(str, metaclass=ExtendedEnumMeta):
    API_KEY = "api_key"
    RESOURCE_PRINCIPAL = "resource_principal"
    INSTANCE_PRINCIPAL = "instance_principal"


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass()
class AuthState(metaclass=SingletonMeta):
    """
    Class stores state of variables specified for auth method, configuration,
    configuration file location, profile name, signer or signer_callable, which
    set by use at any given time and can be provided by this class in any ADS module.
    """

    oci_iam_type: str = None
    oci_cli_auth: str = None
    oci_config_path: str = None
    oci_key_profile: str = None
    oci_config: str = None
    oci_signer: Any = None
    oci_signer_callable: Callable = None
    oci_signer_kwargs: Dict = None
    oci_client_kwargs: Dict = None

    def __post_init__(self):
        self.oci_iam_type = self.oci_iam_type or os.environ.get(
            "OCI_IAM_TYPE", AuthType.API_KEY
        )
        self.oci_cli_auth = self.oci_cli_auth or os.environ.get(
            "OCI_CLI_AUTH", AuthType.RESOURCE_PRINCIPAL
        )
        self.oci_config_path = self.oci_config_path or os.environ.get(
            "OCI_CONFIG_LOCATION", DEFAULT_LOCATION
        )
        self.oci_key_profile = self.oci_key_profile or os.environ.get(
            "OCI_CONFIG_PROFILE", DEFAULT_PROFILE
        )
        self.oci_config = self.oci_config or {}
        self.oci_signer = self.oci_signer
        self.oci_signer_callable = self.oci_signer_callable
        self.oci_signer_kwargs = self.oci_signer_kwargs or {}
        self.oci_client_kwargs = self.oci_client_kwargs or {}


def set_auth(
    auth: Optional[str] = AuthType.API_KEY,
    oci_config_location: Optional[str] = DEFAULT_LOCATION,
    profile: Optional[str] = DEFAULT_PROFILE,
    config: Optional[Dict] = {},
    signer: Optional[Any] = None,
    signer_callable: Optional[Callable] = None,
    signer_kwargs: Optional[Dict] = {},
    client_kwargs: Optional[Dict] = {},
) -> None:
    """
    Save type of authentication, profile, config location, config (keypair identity) or signer, which will be used
    when actual creation of config or signer happens.

    Parameters
    ----------
    auth: Optional[str], default 'api_key'
        'api_key', 'resource_principal' or 'instance_principal'. Enable/disable resource principal identity,
         instance principal or keypair identity in a notebook session
    oci_config_location: Optional[str], default oci.config.DEFAULT_LOCATION, which is '~/.oci/config'
        config file location
    profile: Optional[str], default is DEFAULT_PROFILE, which is 'DEFAULT'
         profile name for api keys config file
    config: Optional[Dict], default {}
        created config dictionary
    signer: Optional[Any], default None
        created signer, can be resource principals signer, instance principal signer or other.
        Check documentation for more signers: https://docs.oracle.com/en-us/iaas/tools/python/latest/api/signing.html
    signer_callable: Optional[Callable], default None
        a callable object that returns signer
    signer_kwargs: Optional[Dict], default None
        parameters accepted by the signer.
        Check documentation: https://docs.oracle.com/en-us/iaas/tools/python/latest/api/signing.html
    client_kwargs: Optional[Dict], default None
        Additional keyword arguments for initializing the OCI client.
        Example: client_kwargs = {"timeout": 60}

    Examples
    --------
    >>> ads.set_auth("api_key") # default signer is set to api keys

    >>> ads.set_auth("api_key", profile = "TEST") # default signer is set to api keys and to use TEST profile

    >>> ads.set_auth("api_key", oci_config_location = "other_config_location") # use non-default oci_config_location

    >>> ads.set_auth("api_key", client_kwargs={"timeout": 60}) # default signer with connection and read timeouts set to 60 seconds for the client.

    >>> other_config = oci.config.from_file("other_config_location", "OTHER_PROFILE") # Create non-default config
    >>> ads.set_auth(config=other_config) # Set api keys type of authentication based on provided config

    >>> ads.set_auth("resource_principal")  # Set resource principal authentication

    >>> ads.set_auth("instance_principal")  # Set instance principal authentication

    >>> singer = oci.auth.signers.get_resource_principals_signer()
    >>> ads.auth.create_signer(config={}, singer=signer) # resource principals authentication dictionary created

    >>> signer_callable = oci.auth.signers.ResourcePrincipalsFederationSigner
    >>> ads.set_auth(signer_callable=signer_callable) # Set resource principal federation singer callable

    >>> signer_callable = oci.auth.signers.InstancePrincipalsSecurityTokenSigner
    >>> signer_kwargs = dict(log_requests=True) # will log the request url and response data when retrieving
    >>> # instance principals authentication dictionary created based on callable with kwargs parameters:
    >>> ads.set_auth(signer_callable=signer_callable, signer_kwargs=signer_kwargs)
    """
    auth_state = AuthState()

    valid_auth_keys = AuthFactory.classes.keys()
    if auth in valid_auth_keys:
        auth_state.oci_iam_type = auth
    else:
        raise ValueError(
            f"Allowed values are: {valid_auth_keys}. If you wish to use other authentication form, "
            f"pass a valid signer or use signer_callable and signer_kwargs"
        )

    if config and (
        oci_config_location != DEFAULT_LOCATION or profile != DEFAULT_PROFILE
    ):
        raise ValueError(
            f"'config' and 'oci_config_location', 'profile' pair are mutually exclusive."
            f"Please specify 'config' OR 'oci_config_location', 'profile' pair."
        )

    auth_state.oci_config = config
    auth_state.oci_key_profile = profile
    if auth == AuthType.API_KEY and not signer and not signer_callable:
        if os.path.exists(os.path.expanduser(oci_config_location)):
            auth_state.oci_config_path = oci_config_location
        else:
            raise ValueError(
                f"{oci_config_location} path does not exist, please provide existing path to config file."
            )

    auth_state.oci_signer = signer
    auth_state.oci_signer_callable = signer_callable
    auth_state.oci_signer_kwargs = signer_kwargs
    auth_state.oci_client_kwargs = client_kwargs


def api_keys(
    oci_config: str = os.path.join(os.path.expanduser("~"), ".oci", "config"),
    profile: str = DEFAULT_PROFILE,
    client_kwargs: Dict = None,
) -> Dict:
    """
    Prepares authentication and extra arguments necessary for creating clients for different OCI services using API
    Keys.

    Parameters
    ----------
    oci_config: Optional[str], default is $HOME/.oci/config
        OCI authentication config file location.
    profile: Optional[str], is DEFAULT_PROFILE, which is 'DEFAULT'
        Profile name to select from the config file.
    client_kwargs: Optional[Dict], default None
        kwargs that are required to instantiate the Client if we need to override the defaults.

    Returns
    -------
    dict
        Contains keys - config, signer and client_kwargs.

        - The config contains the config loaded from the configuration loaded from `oci_config`.
        - The signer contains the signer object created from the api keys.
        - client_kwargs contains the `client_kwargs` that was passed in as input parameter.

    Examples
    --------
    >>> from ads.common import oci_client as oc
    >>> auth = ads.auth.api_keys(oci_config="/home/datascience/.oci/config", profile="TEST", client_kwargs={"timeout": 6000})
    >>> oc.OCIClientFactory(**auth).object_storage # Creates Object storage client with timeout set to 6000 using API Key authentication
    """
    signer_args = dict(
        oci_config_location=oci_config,
        oci_key_profile=profile,
        client_kwargs=client_kwargs,
    )
    signer_generator = AuthFactory().signerGenerator(AuthType.API_KEY)
    return signer_generator(signer_args).create_signer()


def resource_principal(
    client_kwargs: Optional[Dict] = None,
) -> Dict:
    """
    Prepares authentication and extra arguments necessary for creating clients for different OCI services using
    Resource Principals.

    Parameters
    ----------
    client_kwargs: Dict, default None
        kwargs that are required to instantiate the Client if we need to override the defaults.

    Returns
    -------
    dict
        Contains keys - config, signer and client_kwargs.

        - The config contains and empty dictionary.
        - The signer contains the signer object created from the resource principal.
        - client_kwargs contains the `client_kwargs` that was passed in as input parameter.

    Examples
    --------
    >>> from ads.common import oci_client as oc
    >>> auth = ads.auth.resource_principal({"timeout": 6000})
    >>> oc.OCIClientFactory(**auth).object_storage # Creates Object Storage client with timeout set to 6000 seconds using resource principal authentication
    """
    signer_args = dict(client_kwargs=client_kwargs)
    signer_generator = AuthFactory().signerGenerator(AuthType.RESOURCE_PRINCIPAL)
    return signer_generator(signer_args).create_signer()


def create_signer(
    auth_type: Optional[str] = AuthType.API_KEY,
    oci_config_location: Optional[str] = DEFAULT_LOCATION,
    profile: Optional[str] = DEFAULT_PROFILE,
    config: Optional[Dict] = {},
    signer: Optional[Any] = None,
    signer_callable: Optional[Callable] = None,
    signer_kwargs: Optional[Dict] = {},
    client_kwargs: Optional[Dict] = None,
) -> Dict:
    """
    Prepares authentication and extra arguments necessary for creating clients for different OCI services based on
    provided parameters. If `signer` or `signer`_callable` provided, authentication with that signer will be created.
    If `config` provided, api_key type of authentication will be created. Accepted values for `auth_type`:
    `api_key` (default), 'instance_principal', 'resource_principal'.

    Parameters
    ----------
    auth_type: Optional[str], default 'api_key'
        'api_key', 'resource_principal' or 'instance_principal'. Enable/disable resource principal identity,
         instance principal or keypair identity in a notebook session
    oci_config_location: Optional[str], default oci.config.DEFAULT_LOCATION, which is '~/.oci/config'
        config file location
    profile: Optional[str], default is DEFAULT_PROFILE, which is 'DEFAULT'
         profile name for api keys config file
    config: Optional[Dict], default {}
        created config dictionary
    signer: Optional[Any], default None
        created signer, can be resource principals signer, instance principal signer or other.
        Check documentation for more signers: https://docs.oracle.com/en-us/iaas/tools/python/latest/api/signing.html
    signer_callable: Optional[Callable], default None
        a callable object that returns signer
    signer_kwargs: Optional[Dict], default None
        parameters accepted by the signer.
        Check documentation: https://docs.oracle.com/en-us/iaas/tools/python/latest/api/signing.html
    client_kwargs : dict
        kwargs that are required to instantiate the Client if we need to override the defaults

    Examples
    --------
    >>> import ads
    >>> auth = ads.auth.create_signer() # api_key type of authentication dictionary created with default config location and default profile

    >>> config = oci.config.from_file("other_config_location", "OTHER_PROFILE")
    >>> auth = ads.auth.create_signer(config=config) # api_key type of authentication dictionary created based on provided config

    >>> singer = oci.auth.signers.get_resource_principals_signer()
    >>> auth = ads.auth.create_signer(config={}, signer=signer) # resource principals authentication dictionary created

    >>> auth = ads.auth.create_signer(auth_type='instance_principal') # instance principals authentication dictionary created

    >>> signer_callable = oci.auth.signers.InstancePrincipalsSecurityTokenSigner
    >>> signer_kwargs = dict(log_requests=True) # will log the request url and response data when retrieving
    >>> auth = ads.auth.create_signer(signer_callable=signer_callable, signer_kwargs=signer_kwargs) # instance principals authentication dictionary created based on callable with kwargs parameters
    """
    if signer or signer_callable:
        configuration = ads.telemetry.update_oci_client_config()
        if signer_callable:
            signer = signer_callable(**signer_kwargs)
        signer_dict = {
            "config": configuration,
            "signer": signer,
            "client_kwargs": client_kwargs,
        }
        logger.info(f"Using authentication signer type {type(signer)}.")
        return signer_dict
    else:
        signer_args = dict(
            oci_config_location=oci_config_location,
            oci_key_profile=profile,
            oci_config=config,
            client_kwargs=client_kwargs,
        )
        if config:
            auth_type = AuthType.API_KEY

        signer_generator = AuthFactory().signerGenerator(auth_type)

        return signer_generator(signer_args).create_signer()


def default_signer(client_kwargs: Optional[Dict] = None) -> Dict:
    """
    Prepares authentication and extra arguments necessary for creating clients for different OCI services based on
    the default authentication setting for the session. Refer ads.set_auth API for further reference.

    Parameters
    ----------
    client_kwargs : dict
        kwargs that are required to instantiate the Client if we need to override the defaults.
        Example: client_kwargs = {"timeout": 60}

    Returns
    -------
    dict
        Contains keys - config, signer and client_kwargs.

        - The config contains the config loaded from the configuration loaded from the default location if the default
          auth mode is API keys, otherwise it is empty dictionary.
        - The signer contains the signer object created from default auth mode.
        - client_kwargs contains the `client_kwargs` that was passed in as input parameter.

    Examples
    --------
    >>> import ads
    >>> from ads.common import oci_client as oc

    >>> auth = ads.auth.default_signer()
    >>> oc.OCIClientFactory(**auth).object_storage # Creates Object storage client

    >>> ads.set_auth("resource_principal")
    >>> auth = ads.auth.default_signer()
    >>> oc.OCIClientFactory(**auth).object_storage # Creates Object storage client using resource principal authentication

    >>> signer_callable = oci.auth.signers.InstancePrincipalsSecurityTokenSigner
    >>> ads.set_auth(signer_callable=signer_callable) # Set instance principal callable
    >>> auth = ads.auth.default_signer() # signer_callable instantiated
    >>> oc.OCIClientFactory(**auth).object_storage # Creates Object storage client using instance principal authentication
    """
    auth_state = AuthState()
    if auth_state.oci_signer or auth_state.oci_signer_callable:
        configuration = ads.telemetry.update_oci_client_config()
        signer = auth_state.oci_signer
        if auth_state.oci_signer_callable:
            signer_kwargs = auth_state.oci_signer_kwargs or {}
            signer = auth_state.oci_signer_callable(**signer_kwargs)
        signer_dict = {
            "config": configuration,
            "signer": signer,
            "client_kwargs": {
                **(auth_state.oci_client_kwargs or {}),
                **(client_kwargs or {}),
            },
        }
        logger.info(f"Using authentication signer type {type(signer)}.")
        return signer_dict
    else:
        signer_args = dict(
            oci_config_location=auth_state.oci_config_path,
            oci_key_profile=auth_state.oci_key_profile,
            oci_config=auth_state.oci_config,
            client_kwargs={
                **(auth_state.oci_client_kwargs or {}),
                **(client_kwargs or {}),
            },
        )
        signer_generator = AuthFactory().signerGenerator(auth_state.oci_iam_type)
        return signer_generator(signer_args).create_signer()


@deprecated(
    "2.7.3",
    details="Deprecated, use: from ads.common.auth import create_signer. https://accelerated-data-science.readthedocs.io/en/latest/user_guide/cli/authentication.html#overriding-defaults.",
)
def get_signer(
    oci_config: Optional[str] = None, oci_profile: Optional[str] = None, **client_kwargs
) -> Dict:
    """
    Provides config and signer based given parameters. If oci_config (api key config file location) and
    oci_profile specified new signer will ge generated. Else singer of a type specified in OCI_CLI_AUTH
    environment variable will be used to generate signer and return. If OCI_CLI_AUTH not set,
    resource principal signer will be provided. Accepted values for OCI_CLI_AUTH: 'api_key',
    'instance_principal', 'resource_principal'.

    Parameters
    ----------
    oci_config: Optional[str], default None
        Path to the config file
    oci_profile: Optional[str], default None
        the profile to load from the config file
    client_kwargs:
        kwargs that are required to instantiate the Client if we need to override the defaults
    """
    if oci_config and oci_profile:
        signer_args = dict(
            oci_config_location=oci_config,
            oci_key_profile=oci_profile,
            client_kwargs=client_kwargs,
        )
        signer_generator = AuthFactory().signerGenerator(AuthType.API_KEY)
    else:
        oci_cli_auth = (
            AuthState().oci_cli_auth
        )  # "resource_principal", if env variable OCI_CLI_AUTH not set
        valid_auth_keys = AuthFactory.classes.keys()
        if oci_cli_auth not in valid_auth_keys:
            oci_cli_auth = AuthType.RESOURCE_PRINCIPAL
        signer_args = dict(
            client_kwargs=client_kwargs,
        )
        signer_generator = AuthFactory().signerGenerator(oci_cli_auth)
    return signer_generator(signer_args).create_signer()


class AuthSignerGenerator:  # pragma: no cover
    """
    Abstract class for auth configuration and signer creation.
    """

    def create_signer(self):
        pass


class APIKey(AuthSignerGenerator):
    """
    Creates api keys auth instance. This signer is intended to be used when signing requests for
    a given user - it requires that user’s ID, their private key and certificate fingerprint.
    It prepares extra arguments necessary for creating clients for variety of OCI services.
    """

    def __init__(self, args: Optional[Dict] = None):
        """
        Signer created based on args provided. If not provided current values of according arguments
        will be used from current global state from AuthState class.

        Parameters
        ----------
        args: dict
            args that are required to create api key config and signer. Contains keys: oci_config,
            oci_config_location, oci_key_profile, client_kwargs.

            - oci_config is a configuration dict that can be used to create clients
            - oci_config_location - path to config file
            - oci_key_profile - the profile to load from config file
            - client_kwargs - optional parameters for OCI client creation in next steps
        """
        self.oci_config = args.get("oci_config")
        self.oci_config_location = args.get("oci_config_location")
        self.oci_key_profile = args.get("oci_key_profile")
        self.client_kwargs = args.get("client_kwargs")

    def create_signer(self) -> Dict:
        """
        Creates api keys configuration and signer with extra arguments necessary for creating clients.
        Signer constructed from the `oci_config` provided. If not 'oci_config', configuration will be
        constructed from 'oci_config_location' and 'oci_key_profile' in place.

        Resturns
        --------
        dict
            Contains keys - config, signer and client_kwargs.

            - config contains the configuration information
            - signer contains the signer object created. It is instantiated from signer_callable, or
            signer provided in args used, or instantiated in place
            - client_kwargs contains the `client_kwargs` that was passed in as input parameter

        Examples
        --------
        >>> signer_args = dict(
        >>>     client_kwargs=client_kwargs
        >>> )
        >>> signer_generator = AuthFactory().signerGenerator(AuthType.API_KEY)
        >>> signer_generator(signer_args).create_signer()
        """
        if self.oci_config:
            configuration = ads.telemetry.update_oci_client_config(self.oci_config)
        else:
            configuration = ads.telemetry.update_oci_client_config(
                oci.config.from_file(self.oci_config_location, self.oci_key_profile)
            )
        logger.info(f"Using 'api_key' authentication.")
        return {
            "config": configuration,
            "signer": oci.signer.Signer(
                configuration["tenancy"],
                configuration["user"],
                configuration["fingerprint"],
                configuration["key_file"],
                configuration.get("pass_phrase"),
            ),
            "client_kwargs": self.client_kwargs,
        }


class ResourcePrincipal(AuthSignerGenerator):
    """
    Creates Resource Principal signer - a security token for a resource principal.
    It prepares extra arguments necessary for creating clients for variety of OCI services.
    """

    def __init__(self, args: Optional[Dict] = None):
        """
        Signer created based on args provided. If not provided current values of according arguments
        will be used from current global state from AuthState class.

        Parameters
        ----------
        args: dict
            args that are required to create Resource Principal signer. Contains keys: client_kwargs.

            - client_kwargs - optional parameters for OCI client creation in next steps
        """
        self.client_kwargs = args.get("client_kwargs")

    def create_signer(self) -> Dict:
        """
        Creates Resource Principal signer with extra arguments necessary for creating clients.

        Resturns
        --------
        dict
            Contains keys - config, signer and client_kwargs.

            - config contains the configuration information
            - signer contains the signer object created. It is instantiated from signer_callable, or
            signer provided in args used, or instantiated in place
            - client_kwargs contains the `client_kwargs` that was passed in as input parameter

        Examples
        --------
        >>> signer_args = dict(
        >>>     signer=oci.auth.signers.get_resource_principals_signer()
        >>> )
        >>> signer_generator = AuthFactory().signerGenerator(AuthType.RESOURCE_PRINCIPAL)
        >>> signer_generator(signer_args).create_signer()
        """
        configuration = ads.telemetry.update_oci_client_config()
        signer_dict = {
            "config": configuration,
            "signer": oci.auth.signers.get_resource_principals_signer(),
            "client_kwargs": self.client_kwargs,
        }
        logger.info(f"Using 'resource_principal' authentication.")
        return signer_dict


class InstancePrincipal(AuthSignerGenerator):
    """
    Creates Instance Principal signer - a SecurityTokenSigner which uses a security token for an instance
    principal. It prepares extra arguments necessary for creating clients for variety of OCI services.
    """

    def __init__(self, args: Optional[Dict] = None):
        """
        Signer created based on args provided. If not provided current values of according arguments
        will be used from current global state from AuthState class.

        Parameters
        ----------
        args: dict
            args that are required to create Instance Principal signer. Contains keys: signer_kwargs, client_kwargs.

            - signer_kwargs - optional parameters required to instantiate instance principal signer
            - client_kwargs - optional parameters for OCI client creation in next steps
        """
        self.signer_kwargs = args.get("signer_kwargs", dict())
        self.client_kwargs = args.get("client_kwargs")

    def create_signer(self) -> Dict:
        """
        Creates Instance Principal signer with extra arguments necessary for creating clients.
        Signer instantiated from the `signer_callable` or if the `signer` provided is will be return by this method.
        If `signer_callable` or `signer` not provided new signer will be created in place.

        Resturns
        --------
        dict
            Contains keys - config, signer and client_kwargs.

            - config contains the configuration information
            - signer contains the signer object created. It is instantiated from signer_callable, or
            signer provided in args used, or instantiated in place
            - client_kwargs contains the `client_kwargs` that was passed in as input parameter

        Examples
        --------
        >>> signer_args = dict(signer_kwargs={"log_requests": True})
        >>> signer_generator = AuthFactory().signerGenerator(AuthType.INSTANCE_PRINCIPAL)
        >>> signer_generator(signer_args).create_signer()
        """
        configuration = ads.telemetry.update_oci_client_config()
        signer_dict = {
            "config": configuration,
            "signer": oci.auth.signers.InstancePrincipalsSecurityTokenSigner(
                **self.signer_kwargs
            ),
            "client_kwargs": self.client_kwargs,
        }
        logger.info(f"Using 'instance_principal' authentication.")
        return signer_dict


class AuthFactory:
    """
    AuthFactory class which contains list of registered signers and alllows to register new signers.
    Check documentation for more signers: https://docs.oracle.com/en-us/iaas/tools/python/latest/api/signing.html.

    Current signers:
        * APIKey
        * ResourcePrincipal
        * InstancePrincipal
    """

    classes = {
        AuthType.API_KEY: APIKey,
        AuthType.RESOURCE_PRINCIPAL: ResourcePrincipal,
        AuthType.INSTANCE_PRINCIPAL: InstancePrincipal,
    }

    @classmethod
    def register(cls, signer_type: str, signer: Any) -> None:
        """Registers a new signer.

        Parameters
        ----------
        signer_type: str
            Singer type to be registers
        signer: RecordParser
            A new Singer class to be registered.

        Returns
        -------
        None
            Nothing.
        """
        cls.classes[signer_type] = signer

    def signerGenerator(self, iam_type: Optional[str] = "api_key"):
        """
        Generates signer classes based of iam_type, which specify one of auth methods:
        'api_key', 'resource_principal' or 'instance_principal'.

        Parameters
        ----------
        iam_type: str, default 'api_key'
            type of auth provided in IAM_TYPE environment variable or set in parameters in
            ads.set_auth() method.

        Returns
        -------
        :class:`APIKey` or :class:`ResourcePrincipal` or :class:`InstancePrincipal`
            returns one of classes, which implements creation of signer of specified type

        Raises
        ------
        ValueError
            If iam_type is not supported.
        """

        valid_auth_keys = AuthFactory.classes.keys()
        if iam_type in valid_auth_keys:
            return AuthFactory.classes[iam_type]
        else:
            raise ValueError(
                f"Allowed values are: {valid_auth_keys}. If you wish to use other authentication form, "
                f"pass a valid signer or use signer_callable and signer_kwargs"
            )


class OCIAuthContext:
    """
    OCIAuthContext used in 'with' statement for properly managing global authentication type
    and global configuration profile parameters.

    Examples
    --------
    >>> from ads.jobs import DataFlowRun
    >>> with OCIAuthContext(profile='TEST'):
    >>>     df_run = DataFlowRun.from_ocid(run_id)
    """

    @deprecated(
        "2.7.3",
        details="Deprecated, use: from ads.common.auth import AuthContext",
    )
    def __init__(self, profile: str = None):
        """
        Initialize class OCIAuthContext and saves global state of authentication type and configuration profile.

        Parameters
        ----------
        profile: str, default is None
            profile name for api keys config file
        """
        self.profile = profile
        self.prev_mode = AuthState().oci_iam_type
        self.prev_profile = AuthState().oci_key_profile
        self.oci_cli_auth = AuthState().oci_cli_auth

    @deprecated(
        "2.7.3",
        details="Deprecated, use: from ads.common.auth import AuthContext",
    )
    def __enter__(self):
        """
        When called by the 'with' statement and if 'profile' provided - 'api_key' authentication with 'profile' used.
        If 'profile' not provided, authentication method will be 'resource_principal'.
        """
        if self.profile:
            ads.set_auth(auth=AuthType.API_KEY, profile=self.profile)
            logger.info(f"OCI profile set to {self.profile}")
        else:
            ads.set_auth(auth=AuthType.RESOURCE_PRINCIPAL)
            logger.info(f"OCI auth set to resource principal")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        When called by the 'with' statement restores initial state of authentication type and profile value.
        """
        ads.set_auth(auth=self.prev_mode, profile=self.prev_profile)


class AuthContext:
    """
    AuthContext used in 'with' statement for properly managing global authentication type, signer, config
    and global configuration parameters.

    Examples
    --------
    >>> from ads import set_auth
    >>> from ads.jobs import DataFlowRun
    >>> with AuthContext(auth='resource_principal'):
    >>>     df_run = DataFlowRun.from_ocid(run_id)

    >>> from ads.model.framework.sklearn_model import SklearnModel
    >>> model = SklearnModel.from_model_artifact(uri="model_artifact_path", artifact_dir="model_artifact_path")
    >>> set_auth(auth='api_key', oci_config_location="~/.oci/config")
    >>> with AuthContext(auth='api_key', oci_config_location="~/another_config_location/config"):
    >>>     # upload model to Object Storage using config from another_config_location/config
    >>>     model.upload_artifact(uri="oci://bucket@namespace/prefix/")
    >>> # upload model to Object Storage using config from ~/.oci/config, which was set before 'with AuthContext():'
    >>> model.upload_artifact(uri="oci://bucket@namespace/prefix/")
    """

    def __init__(self, **kwargs):
        """
        Initialize class AuthContext and saves global state of authentication type, signer, config
        and global configuration parameters.

        Parameters
        ----------
        **kwargs: optional, list of parameters passed to ads.set_auth() method, which can be:
            auth: Optional[str], default 'api_key'
                'api_key', 'resource_principal' or 'instance_principal'. Enable/disable resource principal
                identity, instance principal or keypair identity
            oci_config_location: Optional[str], default oci.config.DEFAULT_LOCATION, which is '~/.oci/config'
                config file location
            profile: Optional[str], default is DEFAULT_PROFILE, which is 'DEFAULT'
                 profile name for api keys config file
            config: Optional[Dict], default {}
                created config dictionary
            signer: Optional[Any], default None
                created signer, can be resource principals signer, instance principal signer or other
            signer_callable: Optional[Callable], default None
                a callable object that returns signer
            signer_kwargs: Optional[Dict], default None
                parameters accepted by the signer
            client_kwargs: Optional[Dict], default None
                Additional keyword arguments for initializing the OCI client.
                Example: client_kwargs = {"timeout": 60}
        """
        self.kwargs = kwargs

    def __enter__(self):
        """
        When called by the 'with' statement current state of authentication type, signer, config
        and configuration parameters saved.
        """
        self.previous_state = copy.deepcopy(AuthState())
        set_auth(**self.kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        When called by the 'with' statement initial state of authentication type, signer, config
        and configuration parameters restored.
        """
        AuthState().__dict__.update(self.previous_state.__dict__)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import re

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import GISTypedFeature


class LatLonDetector(AbstractTypeDiscoveryDetector):

    _pattern_string = r"^[(]?(\-?\d+\.\d+?),\s*(\-?\d+\.\d+?)[)]?$"

    def __init__(self):
        self.cc = re.compile(LatLonDetector._pattern_string, re.VERBOSE)

    def is_lat_lon(self, name, values):
        return all([self.cc.match(str(x)) for x in values])

    def discover(self, name, series):
        candidates = series.loc[~series.isnull()]

        if self.is_lat_lon(name, candidates.head(1000)):

            logger.debug("column [{}]/[{}] lat/lon".format(name, series.dtype))

            samples = [
                tuple([float(x) for x in self.cc.search(v).groups()])
                for v in candidates.sample(frac=1).head(500).values
            ]
            return GISTypedFeature.build(name, series, samples)

        return False

    @staticmethod
    def extract_x_y(gis_series):
        """takes a GIS series and parses it into a new dataframe with X (longitude) and Y (latitude) columns."""
        cc = re.compile(LatLonDetector._pattern_string, re.VERBOSE)
        lats, lons = zip(*gis_series.dropna().apply(lambda v: cc.search(v).groups()))

        return pd.DataFrame({"X": lons, "Y": lats}).astype(
            {"X": "float64", "Y": "float64"}
        )


if __name__ == "__main__":
    dd = LatLonDetector()

    test_series_1 = [
        "-18.2193965, -93.587285",
        "-21.0255305, -122.478584",
        "85.103913, 19.405744",
        "82.913736, 178.225672",
        "62.9795085, -66.989705",
        "54.5604395, 95.235090",
        "33.970775, -140.939679",
        "40.9680285, -30.369376",
        "51.816119, 175.979008",
        "-48.7882365, 84.035621",
    ]

    test_series_2 = [
        "69.196241,-125.017615",
        "5.2272595,-143.465712",
        "-33.9855425,-153.445155",
        "43.340610,86.460554",
        "24.2811855,-162.380403",
        "2.7849025,-7.328156",
        "45.033805,157.490179",
        "-1.818319,-80.681214",
        "-44.510428,-169.269477",
        "-56.3344375,-166.407038",
    ]

    test_series_3 = ["(54.3392995,-11.801615)"]

    print(dd.discover("test_series_1", pd.Series(test_series_1)))
    print(dd.discover("test_series_2", pd.Series(test_series_2)))
    print(dd.discover("test_series_3", pd.Series(test_series_3)))

    print(LatLonDetector.extract_x_y(pd.Series(test_series_2)))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import PIL

from ads.model.extractor.huggingface_extractor import HuggingFaceExtractor
from ads.model.generic_model import FrameworkSpecificModel
from ads.model.model_properties import ModelProperties
from ads.model.serde.model_serializer import HuggingFaceSerializerType
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.model.serde.model_serializer import ModelSerializerType
from ads.model.serde.common import SERDE


class HuggingFacePipelineModel(FrameworkSpecificModel):
    """HuggingFacePipelineModel class for estimators from HuggingFace framework.

    Attributes
    ----------
    algorithm: str
        The algorithm of the model.
    artifact_dir: str
        Artifact directory to store the files needed for deployment.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained HuggingFace Pipeline using transformers.
    framework: str
        "transformers", the framework name of the model.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
    model_deployment: ModelDeployment
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model.
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples - Image Classification
    -------------------------------
    >>> from transformers import pipeline
    >>> import tempfile
    >>> import PIL.Image
    >>> import ads
    >>> import requests
    >>> import cloudpickle
    >>> ## Download image data
    >>> image_url = "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/pipeline-cat-chonk.jpeg"
    >>> image = PIL.Image.open(requests.get(image_link, stream=True).raw)
    >>> image_bytes = cloudpickle.dumps(image) # convert image to bytes
    >>> ## Download a pretrained model
    >>> vision_classifier = pipeline(model="google/vit-base-patch16-224")
    >>> preds = vision_classifier(images=image)
    >>> ## Initiate a HuggingFacePipelineModel instance
    >>> vision_model = HuggingFacePipelineModel(vision_classifier, artifact_dir=tempfile.mkdtemp())
    >>> ## Prepare
    >>> vision_model.prepare(inference_conda_env="pytorch110_p38_cpu_v1", force_overwrite=True)
    >>> ## Verify
    >>> vision_model.verify(image)
    >>> vision_model.verify(image_bytes)
    >>> ## Save
    >>> vision_model.save()
    >>> ## Deploy
    >>> log_group_id = "<log_group_id>"
    >>> log_id = "<log_id>"
    >>> vision_model.deploy(deployment_bandwidth_mbps=1000,
    ...                wait_for_completion=False,
    ...                deployment_log_group_id = log_group_id,
    ...                deployment_access_log_id = log_id,
    ...                deployment_predict_log_id = log_id)
    >>> ## Predict from endpoint
    >>> vision_model.predict(image)
    >>> vision_model.predict(image_bytes)
    >>> ### Invoke the model
    >>> auth = ads.common.auth.default_signer()['signer']
    >>> endpoint = vision_model.model_deployment.url + "/predict"
    >>> headers = {"Content-Type": "application/octet-stream"}
    >>> requests.post(endpoint, data=image_bytes, auth=auth, headers=headers).json()

    Examples - Image Segmentation
    -----------------------------
    >>> from transformers import pipeline
    >>> import tempfile
    >>> import PIL.Image
    >>> import ads
    >>> import requests
    >>> import cloudpickle
    >>> ## Download image data
    >>> image_url = "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/pipeline-cat-chonk.jpeg"
    >>> image = PIL.Image.open(requests.get(image_link, stream=True).raw)
    >>> image_bytes = cloudpickle.dumps(image) # convert image to bytes
    >>> ## Download pretrained model
    >>> segmenter = pipeline(task="image-segmentation")
    >>> preds = segmenter(image)
    >>> ## Initiate a HuggingFacePipelineModel instance
    >>> segmentation_model = HuggingFacePipelineModel(segmenter, artifact_dir=empfile.mkdtemp())
    >>> ## Prepare
    >>> conda = "oci://bucket@namespace/path/to/conda/pack"
    >>> python_version = "3.8"
    >>> segmentation_model.prepare(inference_conda_env=conda, inference_python_version = python_version, force_overwrite=True)
    >>> ## Verify
    >>> segmentation_model.verify(data=image)
    >>> segmentation_model.verify(data=image_bytes)
    >>> ## Save
    >>> segmentation_model.save()
    >>> log_group_id = "<log_group_id>"
    >>> log_id = "<log_id>"
    >>> ## Deploy
    >>> segmentation_model.deploy(deployment_bandwidth_mbps=1000,
                    wait_for_completion=False,
                    deployment_log_group_id = log_group_id,
                    deployment_access_log_id = log_id,
                    deployment_predict_log_id = log_id)
    >>> ## Predict from endpoint
    >>> segmentation_model.predict(image)
    >>> segmentation_model.predict(image_bytes)
    >>> ## Invoke the model
    >>> auth = ads.common.auth.default_signer()['signer']

    >>> endpoint = segmentation_model.model_deployment.url + "/predict"
    >>> headers = {"Content-Type": "application/octet-stream"}
    >>> requests.post(endpoint, data=image_bytes, auth=auth, headers=headers).json()

    Examples - Zero Shot Image Classification
    -----------------------------------------
    >>> from transformers import pipeline
    >>> import tempfile
    >>> import PIL.Image
    >>> import ads
    >>> import requests
    >>> import cloudpickle
    >>> ## Download the image data
    >>> image_url = "https://huggingface.co/datasets/Narsil/image_dummy/raw/main/parrots.png"
    >>> image = PIL.Image.open(requests.get(image_link, stream=True).raw)
    >>> image_bytes = cloudpickle.dumps(image)
    >>> ## Download a pretrained model
    >>> classifier = pipeline(model="openai/clip-vit-large-patch14")
    >>> classifier(
            images=image,
            candidate_labels=["animals", "humans", "landscape"],
        )
    >>> ## Initiate a HuggingFacePipelineModel instance
    >>> zero_shot_image_classification_model = HuggingFacePipelineModel(classifier, artifact_dir=empfile.mkdtemp())
    >>> conda = "oci://bucket@namespace/path/to/conda/pack"
    >>> python_version = "3.8"
    >>> ## Prepare
    >>> zero_shot_image_classification_model.prepare(inference_conda_env=conda, inference_python_version = python_version, force_overwrite=True)
    >>> data = {"images": image, "candidate_labels": ["animals", "humans", "landscape"]}
    >>> body = cloudpickle.dumps(data) # convert image to bytes
    >>> ## Verify
    >>> zero_shot_image_classification_model.verify(data=data)
    >>> zero_shot_image_classification_model.verify(data=body)
    >>> ## Save
    >>> zero_shot_image_classification_model.save()
    >>> ## Deploy
    >>> log_group_id = "<log_group_id>"
    >>> log_id = "<log_id>"
    >>> zero_shot_image_classification_model.deploy(deployment_bandwidth_mbps=1000,
                    wait_for_completion=False,
                    deployment_log_group_id = log_group_id,
                    deployment_access_log_id = log_id,
                    deployment_predict_log_id = log_id)
    >>> ## Predict from endpoint
    >>> zero_shot_image_classification_model.predict(image)
    >>> zero_shot_image_classification_model.predict(body)
    >>> ### Invoke the model
    >>> auth = ads.common.auth.default_signer()['signer']
    >>> endpoint = zero_shot_image_classification_model.model_deployment.url + "/predict"
    >>> headers = {"Content-Type": "application/octet-stream"}
    >>> requests.post(endpoint, data=body, auth=auth, headers=headers).json()
    """

    _PREFIX = "huggingface"
    model_save_serializer_type = HuggingFaceSerializerType

    @runtime_dependency(
        module="transformers", install_from=OptionalDependency.HUGGINGFACE
    )
    def __init__(
        self,
        estimator: Callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = model_save_serializer_type.HUGGINGFACE,
        model_input_serializer: Optional[SERDE] = ModelSerializerType.CLOUDPICKLE,
        **kwargs,
    ):
        """
        Initiates a HuggingFacePipelineModel instance.

        Parameters
        ----------
        estimator: Callable
            HuggingFacePipeline Model
        artifact_dir: str
            Directory for generate artifact.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
            For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        HuggingFacePipelineModel
            HuggingFacePipelineModel instance.

        Examples
        --------
        >>> from transformers import pipeline
        >>> import tempfile
        >>> import PIL.Image
        >>> import ads
        >>> import requests
        >>> import cloudpickle
        >>> ## download the image
        >>> image_url = "https://huggingface.co/datasets/Narsil/image_dummy/raw/main/parrots.png"
        >>> image = PIL.Image.open(requests.get(image_link, stream=True).raw)
        >>> image_bytes = cloudpickle.dumps(image)
        >>> ## download the pretrained model
        >>> classifier = pipeline(model="openai/clip-vit-large-patch14")
        >>> classifier(
                images=image,
                candidate_labels=["animals", "humans", "landscape"],
            )
        >>> ## Initiate a HuggingFacePipelineModel instance
        >>> zero_shot_image_classification_model = HuggingFacePipelineModel(classifier, artifact_dir=empfile.mkdtemp())
        >>> ## Prepare a model artifact
        >>> conda = "oci://bucket@namespace/path/to/conda/pack"
        >>> python_version = "3.8"
        >>> zero_shot_image_classification_model.prepare(inference_conda_env=conda, inference_python_version = python_version, force_overwrite=True)
        >>> ## Test data
        >>> data = {"images": image, "candidate_labels": ["animals", "humans", "landscape"]}
        >>> body = cloudpickle.dumps(data) # convert image to bytes
        >>> ## Verify
        >>> zero_shot_image_classification_model.verify(data=data)
        >>> zero_shot_image_classification_model.verify(data=body)
        >>> ## Save
        >>> zero_shot_image_classification_model.save()
        >>> ## Deploy
        >>> log_group_id = "<log_group_id>"
        >>> log_id = "<log_id>"
        >>> zero_shot_image_classification_model.deploy(deployment_bandwidth_mbps=100,
                        wait_for_completion=False,
                        deployment_log_group_id = log_group_id,
                        deployment_access_log_id = log_id,
                        deployment_predict_log_id = log_id)
        >>> zero_shot_image_classification_model.predict(image)
        >>> zero_shot_image_classification_model.predict(body)
        >>> ### Invoke the model by sending bytes
        >>> auth = ads.common.auth.default_signer()['signer']
        >>> endpoint = zero_shot_image_classification_model.model_deployment.url + "/predict"
        >>> headers = {"Content-Type": "application/octet-stream"}
        >>> requests.post(endpoint, data=body, auth=auth, headers=headers).json()
        """
        if not isinstance(estimator, transformers.pipelines.base.Pipeline):
            raise TypeError(
                f"{str(type(estimator))} is not supported in HuggingFacePipelineModel."
            )
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = HuggingFaceExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter
        self.task = self.estimator.task
        self._score_args["task"] = self.estimator.task

    def _handle_model_file_name(
        self, as_onnx: bool = False, model_file_name: str = None
    ):
        """
        The artifact directory to store model files.

        Parameters
        ----------
        as_onnx: bool. Defaults to False
            If set as True, it will be ignored as onnx conversion is not supported.
        model_file_name: str
            Will be ignored as huggingface pipeline requires to folder to store the model
            files and those files will be stored at the artifact directory.

        Returns
        -------
        str
            The artifact directory.
        """
        return self.artifact_dir

    def serialize_model(
        self,
        as_onnx: bool = False,
        force_overwrite: bool = False,
        X_sample: Optional[Union[Dict, str, List, PIL.Image.Image]] = None,
        **kwargs,
    ) -> None:
        """
        Serialize and save HuggingFace model using model specific method.

        Parameters
        ----------
        as_onnx: (bool, optional). Defaults to False.
            If set as True, convert into ONNX model.
        force_overwrite: (bool, optional). Defaults to False.
            If set as True, overwrite serialized model if exists.
        X_sample: Union[Dict, str, List, PIL.Image.Image]. Defaults to None.
            A sample of input data that will be used to generate input schema and detect onnx_args.

        Returns
        -------
        None
            Nothing.
        """

        if as_onnx:
            raise NotImplementedError(
                "HuggingFace Pipeline to onnx conversion is not supported."
            )

        super().serialize_model(
            as_onnx=False,
            force_overwrite=force_overwrite,
            X_sample=X_sample,
            **kwargs,
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABCMeta
from typing import List


class AbstractFrameworkSpecBuilder(metaclass=ABCMeta):
    """
    Provides contract for implementing Framework specific Cluster Spec Builder

    In the example of jobs, this class handles adding framework specific environment variables to the job definition.

    NOTE: This class is not invoked while the cluster is running. Only after a call to `ads opctl`.
    """

    def __init__(self, config):
        self.config = config

    def update(self):
        new_env_vars = []
        # Call DaskFramework().build_env_vars
        # FrameworkFactory.get("kind").build_env_vars("spec, framework")
        user_entrypoint = self.config["spec"]["Runtime"]["spec"]["entrypoint"]
        args = self.config["spec"]["Runtime"]["spec"]["args"]

        # TODO:
        # kwargs = self.config["spec"]["Runtime"]["spec"]["kwargs"]

        nanny_port = self.config["spec"]["Framework"]["spec"]["schedulerConfig"][
            "nannyPortRange"
        ]
        worker_port = self.config["spec"]["Framework"]["spec"]["schedulerConfig"][
            "workerPortRange"
        ]
        worker_timeout = self.config["spec"]["Framework"]["spec"]["workerConfig"][
            "timeout"
        ]

        framework = self.config["spec"]["Framework"]["kind"]
        work_dir = self.config["spec"]["Framework"]["spec"]["workDir"]
        ephemeral = self.config["spec"]["Infrastructure"]["spec"]["cluster"][
            "ephemeral"
        ]
        worker_count = self.config["spec"]["Infrastructure"]["spec"]["cluster"][
            "numWorkers"
        ]

        new_env_vars.extend(
            [
                {"name": "OCI__ENTRY_SCRIPT", "value": user_entrypoint},
                {"name": "OCI__ENTRY_SCRIPT_ARGS", "value": args},  # TODO check with QQ
                {
                    "name": "OCI__ENTRY_SCRIPT_KWARGS",
                    "value": None,
                },  # TODO check with QQ
                {"name": "OCI__NANNY_PORT", "value": nanny_port},
                {"name": "OCI__WORKER_PORT", "value": worker_port},
                {"name": "OCI__WORKER_TIMEOUT", "value": worker_timeout},
                {"name": "OCI__WORKER_COUNT", "value": worker_count},
                {"name": "OCI__CLUSTER_TYPE", "value": framework},
                {
                    "name": "OCI__EPHEMERAL",
                    "value": ephemeral,
                },  # TODO: remove from yaml or build out
                {"name": "OCI__LIFE_SPAN", "value": ""},  # TODO: for testing later
                {"name": "OCI__WORK_DIR", "value": work_dir},
            ]
        )

        new_env_vars = self._add_framework_envs(new_env_vars)
        return update_env_vars(self.config, new_env_vars)

    def _add_framework_envs(self, new_env_vars: List):
        # Overwrite this function in your subclass in order to add/subtract environment variables
        return new_env_vars


def update_env_vars(config, env_vars: List):
    """
    env_vars: List, should be formatted as [{"name": "OCI__XXX", "value": YYY},]
    """
    # TODO move this to a class which checks the version, kind, type, etc.
    config["spec"]["Runtime"]["spec"]["environmentVariables"].extend(env_vars)
    return config



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from abc import ABCMeta


class ExtendedEnumMeta(ABCMeta):
    """The helper metaclass to extend functionality of a generic Enum.

    Methods
    -------
    values(cls) -> list:
        Gets the list of class attributes.
    """

    def __contains__(cls, value):
        return value and value.lower() in tuple(value.lower() for value in cls.values())

    def values(cls) -> list:
        """Gets the list of class attributes.

        Returns
        -------
        list
            The list of class values.
        """
        return tuple(
            value for key, value in cls.__dict__.items() if not key.startswith("_")
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import fsspec
import subprocess
import yaml
import json
import re
import docker
from configparser import ConfigParser

from ads.common.auth import create_signer, AuthType, AuthContext
from urllib.parse import urlparse
from ads.jobs import Job, DataScienceJobRun
from ads.jobs.builders.runtimes.artifact import Artifact
from ads.opctl.utils import publish_image as publish_image_cmd
from ads.opctl.utils import run_command

ini_file = "config.ini"


def _artifact_file_name(cluster_type, version):
    return f"{cluster_type}_{version}.tar.gz".replace("-", "_")


def initialize_workspace(cluster_type, version):
    base_uri = None
    local_file_name = f".tmp_{_artifact_file_name(cluster_type, version)}".replace(
        "-", "_"
    )
    if os.environ.get("OCI_DISTRIBUTED_DEV_MODE"):
        base_uri = dev_mode_base_uri(cluster_type, version)
    else:
        base_uri = production_mode_base_uri(cluster_type, version)
    artificat_location = os.path.join(
        base_uri, _artifact_file_name(cluster_type, version)
    )
    print(f"Downloading from {artificat_location} to {local_file_name}")
    Artifact.copy_from_uri(uri=artificat_location, to_path=local_file_name)
    os.makedirs("oci_dist_training_artifacts", exist_ok=True)
    try:
        subprocess.call(
            ["tar", "-xvf", local_file_name, "-C", "oci_dist_training_artifacts"]
        )
        print(
            f"Check oci_dist_training_artifacts/{cluster_type.replace('-', '_')}/{version}/README.md for build instructions "
        )
    finally:
        if os.path.exists(local_file_name):
            os.remove(local_file_name)


def dev_mode_base_uri(cluster_type, version):
    """
    Temporary method to mock initialization using bucket.
    """
    bucket_name = os.environ.get("OCI_DISTRIBUTED_DEV_MODE_BUCKET")
    namespace = os.environ.get("OCI_DISTRIBUTED_DEV_MODE_NAMESPACE")
    prefix = os.environ.get("OCI_DISTRIBUTED_DEV_MODE_PREFIX")
    base_uri = f"oci://{bucket_name}@{namespace}/{prefix}/"
    return base_uri


def production_mode_base_uri(cluster_type, version):
    return "https://raw.githubusercontent.com/oracle-samples/oci-data-science-ai-samples/master/distributed_training/artifacts/"


def cancel_distributed_run(job_id, cluster_file_name, **kwargs):
    if cluster_file_name:
        with open(cluster_file_name) as cf:
            content = yaml.load(cf, yaml.SafeLoader)
            job_id = content["jobId"]

    with AuthContext():
        import ads

        ads.set_auth(auth="api_key", profile=kwargs.get("oci_profile", "DEFAULT"))

        job = Job.from_datascience_job(job_id)
        runs = job.run_list()
        for job_run in runs:
            print(f"Cancelling Job Run: {job_run.id}")
            try:
                job_run.cancel()
            except Exception as e:
                print(f"Error cancelling: {e}")
        if not runs:
            print(f"No Job runs found for : {job_id}")


def show_config_info(job_id, work_dir, cluster_file_name, worker_info, **kwargs):
    if cluster_file_name:
        with open(cluster_file_name) as cf:
            content = yaml.load(cf, yaml.SafeLoader)
            job_id = content["jobId"]
            work_dir = content["workDir"]
    config_location = os.path.join(work_dir, job_id)
    scheme = urlparse(
        config_location,
    ).scheme

    oci_auth = (
        create_signer(
            AuthType.API_KEY,
            kwargs.get("oci_config") or os.path.expanduser("~/.oci/config"),
            kwargs.get("oci_profile") or "DEFAULT",
        )
        if scheme == "oci"
        else {}
    )
    filesystem = fsspec.filesystem(
        scheme, **oci_auth
    )  # FileSystem class corresponding to the URI scheme.
    main_file = os.path.join(config_location, "MAIN_config.json")
    if filesystem.exists(main_file):
        with fsspec.open(main_file, **oci_auth) as mfile:
            print("Main Info:")
            print(yaml.dump(json.loads(mfile.read())))

        if worker_info:
            files = fsspec.open_files(
                os.path.join(config_location, "WORKER*.json"), **oci_auth
            )
            for fl in files:
                with fsspec.open(
                    f"{'oci://' if oci_auth else ''}{fl.path}", **oci_auth
                ) as f:
                    print(f"Worker Info from {os.path.basename(fl.path)}")
                    print(yaml.dump(json.loads(f.read())))
    else:
        print(
            f"MAIN_config file not found at location {config_location}. If you just started the cluster please wait until the main node is in `inprogress` state"
        )


def verify_image(img_name):
    """
    Verify if the input image exists in OCI registry

    Parameters
    ----------
    img_name
        Name of the docker image

    Returns
        None
    -------

    """
    client = docker.from_env()
    try:
        client.images.get_registry_data(img_name)
        return 1
    except Exception:
        return 0


def update_ini(tag, registry, dockerfile, source_folder, config, nobuild):
    """
    Stores and updates input args in config.ini in Local

    Parameters
    ----------
    tag
        tag of image
    registry
        registry to push to
    dockerfile
        relative path to Dockerfile
    source_folder
        source scripts folder that will be mounted during a local run
    config
        Job config
    nobuild
        flag for building the docker

    Returns
        dict of config.ini file
    -------

    """
    ini = ConfigParser(allow_no_value=True)
    if config is None:
        img_name = "@test"
    else:
        img_name = (
            config.get("spec", {}).get("cluster", {}).get("spec", {}).get("image")
        )

    if tag is not None:
        tag1 = tag
    else:
        if img_name.startswith("@"):
            tag1 = tag
        else:
            if len(img_name.rsplit(":", 1)) == 1:
                tag1 = "latest"
            else:
                tag1 = img_name.rsplit(":", 1)[1]
    if registry is not None:
        registry1 = registry
    else:
        if img_name.startswith("@"):
            registry1 = registry
        else:
            registry1 = img_name.rsplit(":", 1)[0]

    if os.path.isfile(ini_file):
        ini.read(ini_file)
        if tag1 is not None:
            ini.set("main", "tag", tag1)
        if registry1 is not None:
            ini.set("main", "registry", registry1)
        if dockerfile is not None:
            ini.set("main", "dockerfile", dockerfile)
        if source_folder is not None:
            ini.set("main", "source_folder", source_folder)
    else:
        ini.add_section("main")
        if tag1 is not None:
            ini.set("main", "tag", tag1)
        else:
            raise ValueError("tag arg is missing")
        if registry1 is not None:
            ini.set("main", "registry", registry1)
        else:
            raise ValueError("registry arg is missing")
        if dockerfile is not None:
            ini.set("main", "dockerfile", dockerfile)
        else:
            if nobuild:
                ini.set("main", "dockerfile", "DUMMY_PATH")
            else:
                raise ValueError("dockerfile arg is missing")
        if source_folder is not None:
            ini.set("main", "source_folder", source_folder)
        else:
            ini.set("main", "source_folder", ".")
    ini.set("main", "; mount oci keys for local testing", None)
    if ini.has_option("main", "oci_key_mnt"):
        pass
    else:
        ini.set("main", "oci_key_mnt", "~/.oci:/home/oci_dist_training/.oci")
    if ini.get("main", "dockerfile") == "DUMMY_PATH" and not nobuild:
        raise ValueError("dockerfile arg is missing")
    if os.path.exists(ini_file):
        os.remove(ini_file)
    with open(ini_file, "w") as f:
        ini.write(f)
    return ini


def load_ini():
    """
    Loads config.ini from local in dict
    Returns
        None
    -------

    """
    ini = ConfigParser(allow_no_value=True)
    if os.path.isfile(ini_file):
        ini.read(ini_file)
        return ini
    else:
        raise RuntimeError(f"ini_file file {ini_file} not found !")


def increment_tag_in_ini(ini):
    """
    increments tag of image and update latest in config.ini file

    Parameters
    ----------
    ini
        config.ini file dictionary
    Returns
        updated config.ini file
    -------

    """
    ini = increment_tag(ini)
    refresh_ini(ini, "config.ini")
    return ini


def increment_tag(ini):
    """
    Increments the tag of the image

    Parameters
    ----------
    ini
        config.ini file dictionary

    Returns
        updated ini dict
    -------

    """
    tag_name = ini.get("main", "tag")
    m = re.search(r"_v_\d+$", tag_name)
    if m is not None:
        updated_tag = re.sub(
            r"[0-9]+$",
            lambda x: f"{str(int(x.group()) + 1).zfill(len(x.group()))}",
            tag_name,
        )
    else:
        updated_tag = f"{tag_name}_v_1"
    ini.set("main", "tag", updated_tag)
    return ini


def refresh_ini(ini, ini_file):
    """
    write updated config.ini in local

    Parameters
    ----------
    ini
        config.ini file dictionary
    ini_file
        Name of the config file

    Returns
    -------

    """
    if os.path.exists(ini_file):
        os.remove(ini_file)
    with open(ini_file, "w") as f:
        ini.write(f)
    return ini


def docker_build_cmd(ini):
    """
    Builds the docker image

    Parameters
    ----------
    ini
        config.ini file dictionary

    Returns
        None
    -------

    """
    cmd = get_cmd(ini)
    return run_cmd(cmd)


def get_cmd(ini):
    """
    Docker build command

    Parameters
    ----------
    ini
        config.ini file dictionary
    Returns
        Docker build command
    -------

    """
    command = [
        "docker",
        "build",
        "--build-arg",
        "CODE_DIR=" + ini.get("main", "source_folder"),
        "-t",
        ini.get("main", "registry") + ":" + ini.get("main", "tag"),
        "-f",
        ini.get("main", "dockerfile"),
    ]
    if os.environ.get("no_proxy"):
        command += ["--build-arg", f"no_proxy={os.environ['no_proxy']}"]
    if os.environ.get("http_proxy"):
        command += ["--build-arg", f"http_proxy={os.environ['http_proxy']}"]
    if os.environ.get("https_proxy"):
        command += ["--build-arg", f"https_proxy={os.environ['https_proxy']}"]
    command += ["."]
    return command


def run_cmd(cmd):
    """
    Run cli command
    Parameters
    ----------
    cmd
        input cli command

    Returns
        None
    -------

    """
    proc = run_command(cmd)
    if proc.returncode != 0:
        raise RuntimeError(f"Docker build failed.")
    return 1


def horovod_cmd(code_mount, oci_key_mount, config):
    """
    CLI command to run horovod distributed framework on local

    Parameters
    ----------
    code_mount
        source_folder to be mounted
    oci_key_mount
        oci keys to be mounted
    config
        Job config

    Returns
        List: CLI command
    -------

    """
    command = [
        "docker",
        "run",
        "-v",
        code_mount,
        "-v",
        oci_key_mount,
        "--env",
        "OCI_IAM_TYPE=api_key",
        "--rm",
        "--entrypoint",
        "/miniconda/envs/env/bin/horovodrun",
        config["spec"]["cluster"]["spec"]["image"],
        "--gloo",
        "-np",
        "2",
        "-H",
        "localhost:2",
        "/miniconda/envs/env/bin/python",
        config["spec"]["runtime"]["spec"]["entryPoint"],
    ]
    return command


def pytorch_cmd(code_mount, oci_key_mount, config):
    """
    CLI command to run Pytorch distributed framework on local

    Parameters
    ----------
    code_mount
        source_folder to be mounted
    oci_key_mount
        oci keys to be mounted
    config
        Job config

    Returns
        List: CLI command
    -------

    """
    command = [
        "docker",
        "run",
        "-v",
        code_mount,
        "-v",
        oci_key_mount,
        "--env",
        "OCI_IAM_TYPE=api_key",
        "--env",
        "WORLD_SIZE=1",
        "--env",
        "RANK=0",
        "--env",
        "LOCAL_RANK=0",
        "--rm",
        "--entrypoint",
        "/opt/conda/bin/python",
        config["spec"]["cluster"]["spec"]["image"],
        config["spec"]["runtime"]["spec"]["entryPoint"],
    ]
    return command


def dask_cmd(code_mount, oci_key_mount, config):
    """
    CLI command to run Dask distributed framework on local

    Parameters
    ----------
    code_mount
        source_folder to be mounted
    oci_key_mount
        oci keys to be mounted
    config
        Job config

    Returns
        List: CLI command
    -------

    """
    command = [
        "docker",
        "run",
        "-v",
        code_mount,
        "-v",
        oci_key_mount,
        "--env",
        "OCI_IAM_TYPE=api_key",
        "--env",
        "SCHEDULER_IP=tcp://127.0.0.1",
        "--rm",
        "--entrypoint",
        "/bin/sh",
        config["spec"]["cluster"]["spec"]["image"],
        "-c",
        "(nohup dask-scheduler >scheduler.log &) && (nohup dask-worker localhost:8786 >worker.log &) && "
        "/miniconda/envs/daskenv/bin/python "
        + config["spec"]["runtime"]["spec"]["entryPoint"],
    ]
    return command


def tensorflow_cmd(code_mount, oci_key_mount, config):
    """
    CLI command to run Tensorflow distributed framework on local

    Parameters
    ----------
    code_mount
        source_folder to be mounted
    oci_key_mount
        oci keys to be mounted
    config
        Job config

    Returns
        List: CLI command
    -------

    """
    if "ps" in config["spec"]["cluster"]["spec"]:
        # this is for the parameter server strategy.
        command = [
            "docker",
            "run",
            "-v",
            code_mount,
            "-v",
            oci_key_mount,
            "--env",
            "OCI_IAM_TYPE=api_key",
            "--rm",
            "--entrypoint",
            "/bin/sh",
            config["spec"]["cluster"]["spec"]["image"],
            "/etc/datascience/local_test.sh",
            config["spec"]["runtime"]["spec"]["entryPoint"],
        ]

    else:
        # this is for the mirror and multiworkermirror strategy.
        command = [
            "docker",
            "run",
            "-v",
            code_mount,
            "-v",
            oci_key_mount,
            "--env",
            "OCI_IAM_TYPE=api_key",
            "--env",
            'TF_CONFIG={"cluster": {"worker": ["localhost:12345"]}, "task": {"type": "worker", "index": 0}}',
            "--rm",
            "--entrypoint",
            "/miniconda/bin/python",
            config["spec"]["cluster"]["spec"]["image"],
            config["spec"]["runtime"]["spec"]["entryPoint"],
        ]
    return command


def local_run(config, ini):
    """

    Local run distributed framework on local based on input args

    Parameters
    ----------
    ini
        config.ini file dictionary
    config
        Job run config

    Returns
        None
    -------

    """
    cwd = os.path.join(os.getcwd(), ini.get("main", "source_folder"))
    code_mount = os.path.join(cwd, ":/code/")
    oci_key_mount = (
        os.path.expanduser(ini.get("main", "oci_key_mnt").split(":")[0])
        + ":"
        + ini.get("main", "oci_key_mnt").split(":")[1]
    )
    if config["spec"]["cluster"]["kind"].lower() == "horovod":
        command = horovod_cmd(code_mount, oci_key_mount, config)
    elif config["spec"]["cluster"]["kind"].lower() == "pytorch":
        command = pytorch_cmd(code_mount, oci_key_mount, config)
    elif config["spec"]["cluster"]["kind"].lower() == "dask":
        command = dask_cmd(code_mount, oci_key_mount, config)
    elif config["spec"]["cluster"]["kind"].lower() == "tensorflow":
        command = tensorflow_cmd(code_mount, oci_key_mount, config)
    else:
        raise RuntimeError(f"Framework not supported")
    try:
        command += [str(arg) for arg in config["spec"]["runtime"]["spec"]["args"]]
    except KeyError:
        pass
    print("Running: ", " ".join(command))
    proc = run_command(command)
    if proc.returncode != 0:
        raise RuntimeError(f"Failed to run local")
    return 1


def update_image(config, ini):
    """
    Update Image name in job config with latest build image

    Parameters
    ----------
    config
        Job run config
    ini
        config.ini file dictionary

    Returns
        Updated Job run config
    -------

    """
    config["spec"]["cluster"]["spec"]["image"] = (
        ini.get("main", "registry") + ":" + ini.get("main", "tag")
    )
    return config


def verify_and_publish_image(nopush, config):
    """
    verify existence of docker image in OCI registry and publish it
    Parameters
    ----------
    nopush
        Doesn't push image to OCI registry
    config
        Job run config

    Returns
        None
    -------

    """
    if not nopush:
        publish_image_cmd(config["spec"]["cluster"]["spec"]["image"])
    else:
        if not verify_image(config["spec"]["cluster"]["spec"]["image"]):
            print(
                "\u26A0 Image: "
                + config["spec"]["cluster"]["spec"]["image"]
                + " does not exist in registry"
            )
            print("In order to push the image to registry enter Y else N ")
            inp = input("[Y/N]\n")
            if inp == "Y":
                print("\u2705 pushing image to registry")
                publish_image_cmd(config["spec"]["cluster"]["spec"]["image"])
            else:
                raise RuntimeError(
                    "Stopping the execution as image doesn't exist in OCI registry"
                )
    return 1


def update_config_image(config):
    """
    updates image name in config

    Parameters
    ----------
    config
        Job config

    Returns
        updated config dict
    -------

    """
    ini = ConfigParser(allow_no_value=True)
    img_name = config.get("spec", {}).get("cluster", {}).get("spec", {}).get("image")
    if img_name.startswith("@"):
        if os.path.isfile(ini_file):
            ini.read(ini_file)
            return update_image(config, ini)
        else:
            raise ValueError(
                f"Invalid image name: {img_name} and also not able to locate config.ini file. "
                f"Please update image name in Job config."
            )
    else:
        return config



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.feature_selection import f_regression, f_classif, SelectKBest

from ads.type_discovery.typed_feature import ContinuousTypedFeature
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class FeatureImportance:
    @staticmethod
    def _get_feature_ranking(sampled_df, target, target_type, score_func=None, k=None):
        if score_func is None:
            score_func = (
                f_regression
                if isinstance(target_type, ContinuousTypedFeature)
                else f_classif
            )
        # if there are any Nans, get the most common value from each column and replace. If entire column is Nan and not categorical, replace with 0
        if sampled_df.isnull().any().sum() > 0:
            mode_values = sampled_df.mode().iloc[0]
            for i, col in enumerate(sampled_df.columns):
                if pd.isna(mode_values[i]):
                    if sampled_df[col].dtype.name != "category":
                        mode_values[i] = 0
            sampled_df.fillna(mode_values, inplace=True)

        # label encoding of categorical features
        categorical_columns = sampled_df.select_dtypes(include=["object", "category"])
        if len(categorical_columns.columns) > 0:
            categorical_columns_encoded = categorical_columns.apply(
                preprocessing.LabelEncoder().fit_transform
            )
            sampled_df[categorical_columns.columns.values] = categorical_columns_encoded

        # convert datetimes to float
        datetime_columns = sampled_df.select_dtypes(include=["datetime64[ns]"])
        if len(datetime_columns.columns) > 0:
            sampled_df[
                datetime_columns.columns.values
            ] = datetime_columns.values.astype("float")

        timestamp_columns = sampled_df.select_dtypes(include=["datetime64[ns, UTC]"])
        if len(timestamp_columns.columns) > 0:
            sampled_df[
                timestamp_columns.columns.values
            ] = timestamp_columns.values.astype("datetime64[ns]").astype("float")

        extracter = SelectKBest(score_func=score_func, k="all")
        columns = sampled_df.columns.values
        feature_names = np.delete(columns, np.argwhere(columns == target))
        feature_scores = extracter.fit(
            sampled_df.drop(target, axis=1), sampled_df[target]
        )
        feature_ranking_df = (
            pd.DataFrame(
                sorted(
                    zip(
                        feature_scores.scores_,
                        feature_names[extracter.get_support(indices=True)],
                    ),
                    reverse=True,
                ),
                columns=["scores", "features"],
            )[["features", "scores"]]
            .fillna(0)
            .sort_values(by=["scores"], ascending=False)[:k]
        )
        return feature_ranking_df

    def __init__(self, ds, score_func=None, n=None):
        self.feature_ranking = self._get_feature_ranking(
            ds.sampled_df, ds.type_of_target(), score_func=score_func, k=n
        )

    def __repr__(self):
        return str(self.feature_ranking)

    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def show_in_notebook(self, fig_size=(10, 10)):
        """
        Shows selected features in the notebook with matplotlib.
        """

        if not utils.is_notebook():
            print("show_in_notebook called but not in notebook environment")
            return

        with seaborn.axes_style(style="whitegrid"):
            plt.figure(figsize=fig_size)
            seaborn.barplot(
                x="scores",
                y="features",
                data=self.feature_ranking,
                orient="h",
                palette="Blues_d",
            )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from collections import defaultdict
from configparser import ConfigParser
from copy import copy
from enum import Enum
from typing import Callable, Dict, List, Optional, Tuple, Union, Any

import fsspec
import yaml
from ads.common import auth as authutil
from ads.common.decorator.argument_to_case import ArgumentCase, argument_to_case

try:
    from yaml import CSafeDumper as dumper
except:
    from yaml import SafeDumper as dumper


DEFAULT_CONFIG_PROFILE = "DEFAULT"
DEFAULT_CONFIG_PATH = "~/.ads/config"


class EventType(Enum):
    CHANGE = "change"


class Mode:
    READ = "r"
    WRITE = "w"


class Eventing:
    """The class helper to register event handlers."""

    def __init__(self):
        self._events = defaultdict(set)

    def trigger(self, event: str) -> None:
        """Triggers all the registered callbacks for the particular event."""
        for callback in self._events[event]:
            callback()

    def on(self, event_name: str, callback: Callable) -> None:
        """Registers a callback for the particular event."""
        self._events[event_name].add(callback)


class ConfigSection:
    """The class representing a config section."""

    def __init__(self):
        """Initializes the config section instance."""
        self.events = Eventing()
        self._info = {}

    def clear(self) -> None:
        """Clears the config section values.

        Returns
        -------
        None
            Nothing
        """
        self._info = {}
        self.events.trigger(EventType.CHANGE.value)

    def copy(self) -> "ConfigSection":
        """Makes a copy of a config section.

        Returns
        -------
        ConfigSection
            The instance of a copied ConfigSection.
        """
        return self.__class__()._with_dict(info=copy(self._info), replace=True)

    def _with_dict(
        self, info: Dict[str, Any], replace: Optional[bool] = False
    ) -> "ConfigSection":
        """Populates the config section from a dictionary.

        Parameters
        ----------
        info: Dict[str, Any]
            The config section information in a dictionary format.
        replace: (bool, optional). Defaults to False.
            If set as True, overwrites config section with the new information.

        Returns
        -------
        ConfigSection
            The instance of a ConfigSection.

        Raises
        -----
        TypeError
            If input data is not a dictionary.
        ValueError
            If config section is already contain provided fields
            and `replace` flag set to False.
        """

        if not isinstance(info, dict):
            raise TypeError("The `info` must be a dictionary.")

        common_keys = list(
            set(self._info.keys()).intersection(set(list(map(str.lower, info.keys()))))
        )

        if common_keys and not replace:
            raise ValueError(
                f"The config section is already contain fields: {common_keys}. "
                "Use `replace=True` to overwrite."
            )

        for k, v in info.items():
            self._set(key=k.lower(), value=v, replace=replace)

        return self

    def with_dict(
        self, info: Dict[str, Any], replace: Optional[bool] = False
    ) -> "ConfigSection":
        """Populates the config section from a dictionary.

        Parameters
        ----------
        info: Dict[str, Any]
            The config section information in a dictionary format.
        replace: (bool, optional). Defaults to False.
            If set as True, overwrites config section with the new information.
        """
        self._with_dict(info=info, replace=replace)
        self.events.trigger(EventType.CHANGE.value)
        return self

    def keys(self) -> Tuple[str]:
        """Gets the list of the keys of a config section.

        Returns
        -------
        Tuple[str]
            The list of config section keys.
        """
        return tuple(self._info.keys())

    def to_dict(self) -> Dict[str, Any]:
        """Converts config section to a dictionary.

        Returns
        -------
        Dict[str, Any]
            The config section in a dictionary format.
        """
        return self._info

    @argument_to_case(case=ArgumentCase.LOWER, arguments=["key"])
    def get(self, key: str) -> str:
        """Gets the config section value by key.

        Returns
        -------
        str
            A specific config section value.
        """
        return self._info.get(key)

    @argument_to_case(case=ArgumentCase.LOWER, arguments=["key"])
    def _set(self, key: str, value: str, replace: Optional[bool] = False) -> None:
        """Sets the config section value by key.

        Parameters
        ----------
        key: str
            The config section field key.
        value: str
            The config section field value.

        Returns
        -------
        None
            Nothing

        Raises
        ------
        ValueError
            In case when field with provided key already exists and
            `replace` flag set to False.
        """
        if self._info.get(key) == value:
            return

        if key in self._info and not replace:
            raise ValueError(
                f"The field with key `{key}` already exists. "
                "Use `replace=True` to overwrite."
            )

        self._info[key] = value

    def set(self, key: str, value: str, replace: Optional[bool] = False) -> None:
        """Sets the config section value by key.

        Parameters
        ----------
        key: str
            The config section field key.
        value: str
            The config section field value.

        Returns
        -------
        None
            Nothing
        """
        self._set(key, value, replace)
        self.events.trigger(EventType.CHANGE.value)

    @argument_to_case(case=ArgumentCase.LOWER, arguments=["key"])
    def remove(self, key: str) -> None:
        """Removes the config section field by key.

        Parameters
        ----------
        key: str
            The config section field key.

        Returns
        -------
        None
            Nothing
        """
        self._info.pop(key, None)
        self.events.trigger(EventType.CHANGE.value)

    def __getitem__(self, key: str):
        return self.get(key)

    def __setitem__(self, key: str, value: str):
        self.set(key=key, value=value, replace=True)

    def __bool__(self):
        return any(self._info.values())

    def __repr__(self):
        return yaml.dump(self.to_dict(), Dumper=dumper)


class Config:
    """The class representing a config."""

    __DEFAULT_SECTIONS = {
        DEFAULT_CONFIG_PROFILE: ConfigSection,
    }

    def __init__(
        self,
        uri: Optional[str] = DEFAULT_CONFIG_PATH,
        auth: Optional[Dict] = None,
    ):
        """Initializes a config instance.

        Parameters
        ----------
        uri: (str, optional). Defaults to `~/.ads/config`.
            The path to the config file. Can be local or Object Storage file.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        """
        self._config = {}
        self.auth = auth or authutil.default_signer()

        # configure default config sections
        for key, default_section in self.__DEFAULT_SECTIONS.items():
            self._config[key] = default_section()
            self._config[key].events.on(EventType.CHANGE.value, self._on_change)

        self.uri = os.path.expanduser(uri)
        self._config_parser = ExtendedConfigParser(uri=self.uri, auth=self.auth)

    def _on_change(self):
        pass

    def default(self) -> ConfigSection:
        """Gets default config section.

        Returns
        -------
        ConfigSection
            A default config section.
        """
        return self.section_get(DEFAULT_CONFIG_PROFILE)

    @argument_to_case(case=ArgumentCase.UPPER, arguments=["key"])
    def section_exists(self, key: str) -> bool:
        """Checks if a config section exists.

        Parameters
        ----------
        key: str
            A key of a config section.

        Returns
        -------
        bool
            True if a config section exists, Fasle otherwise.
        """
        return key in self._config

    @argument_to_case(case=ArgumentCase.UPPER, arguments=["key"])
    def section_get(self, key: str) -> ConfigSection:
        """Gets the config section by key.

        Returns
        -------
        ConfigSection
            A config section object.

        Raises
        ------
        KeyError
            If a config section not exists.
        """
        if key not in self._config:
            raise KeyError(f"The config section `{key}` not found.")
        return self._config.get(key)

    @argument_to_case(case=ArgumentCase.UPPER, arguments=["key"])
    def section_set(
        self,
        key: str,
        info: Union[dict, ConfigSection],
        replace: Optional[bool] = False,
    ) -> ConfigSection:
        """
        Sets a config section to config.
        The new config section will be added in case if it doesn't exist.
        Otherwise the existing config section will be merged with the new fields.

        Paramaters
        ----------
        key: str
            A key of a config section.
        info: Union[dict, ConfigSection]
            The config section information in a dictionary or ConfigSection format.
        replace: (bool, optional). Defaults to False.
            If set as True, overwrites config section with the new information.

        Returns
        -------
        ConfigSection
            A config section object.

        Raises
        ------
        ValueError
            If section with goven key is already exist and `replace` flag set to False.
        TypeError
            If input `info` has a wrong format.
        """
        if key in self._config and not replace:
            raise ValueError(
                f"A config section `{key}` is already exist. "
                "Use `replace=True` if you want to overwrite."
            )

        if not isinstance(info, (dict, ConfigSection)):
            raise TypeError(
                "Parameter `info` must be either a `dictionary` or `ConfigSection` object."
            )

        if key not in self._config:
            self._config[key] = ConfigSection()
            self._config[key].events.on(EventType.CHANGE.value, self._on_change)

        if isinstance(info, ConfigSection):
            self._config[key].with_dict(info.copy().to_dict(), replace=replace)
        else:
            self._config[key].with_dict(copy(info), replace=replace)

        return self._config[key]

    @argument_to_case(case=ArgumentCase.UPPER, arguments=["key"])
    def section_remove(self, key: str) -> None:
        """Removes config section form config.

        Parameters
        ----------
        key: str
            A key of a config section that needs to be removed.

        Returns
        -------
        None
            Nothing
        """
        self._config.pop(key, None)
        self._on_change()

    def save(self) -> None:
        """Saves config data to a config file.

        Returns
        -------
        None
            Nothing
        """
        self._config_parser.with_dict(self.to_dict()).save()

    def load(self) -> "Config":
        """Loads config data from a config file.

        Returns
        -------
        Config
            A config object.
        """
        return self.with_dict(self._config_parser.read().to_dict())

    def with_dict(
        self,
        info: Dict[str, Union[Dict[str, Any], ConfigSection]],
        replace: Optional[bool] = False,
    ) -> "Config":
        """Merging dictionary to config.

        Parameters
        ----------
        info: Dict[str, Union[Dict[str, Any], ConfigSection]]
            A dictionary that needs to be merged to config.
        replace: (bool, optional). Defaults to False.
            If set as True, overwrites config section with the new information.

        Returns
        -------
        Config
            A config object.
        """

        self._validate(info)
        try:
            for key, value in info.items():
                self.section_set(key, value, replace=replace)
        finally:
            self._on_change()

        return self

    def to_dict(self) -> Dict[str, Dict[str, Any]]:
        """Converts config to a dictionary format.

        Returns
        -------
        Dict[str, Dict[str, Any]]
            A config in a dictionary format.
        """
        return {key: value.to_dict() for key, value in self._config.items()}

    def keys(self) -> List[str]:
        """Gets the all registered config section keys.

        Returns
        -------
        List[str]
            The list of the all registered config section keys.
        """
        return self._config.keys()

    def _validate(self, info: Dict[str, Union[Dict[str, Any], ConfigSection]]) -> None:
        """Validates input dictionary."""
        if not info or not isinstance(info, Dict):
            raise TypeError("The input data should be a dictionary.")
        for key, value in info.items():
            if value and not isinstance(value, (Dict, ConfigSection)):
                raise ValueError(
                    f"The `{key}` must be a dictionary or a `ConfigSection` instance."
                )

    def __getitem__(self, key: str):
        return self.section_get(key)

    def __setitem__(self, key, value: Union[Dict, ConfigSection]):
        self.section_set(key, value, replace=True)

    def __repr__(self):
        return yaml.dump(self.to_dict(), Dumper=dumper)


class ExtendedConfigParser(ConfigParser):
    """Class helper to read/write information to the config file."""

    def __init__(
        self, uri: Optional[str] = DEFAULT_CONFIG_PATH, auth: Optional[Dict] = None
    ):
        """Initializes a config parser instance.

        Parameters
        ----------
        uri: (str, optional). Defaults to `~/.ads/config`.
            The path to the config file. Can be local or Object Storage file.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        """
        super().__init__(default_section="EXCLUDE_DEFAULT_SECTION")
        self.auth = auth or authutil.default_signer()
        self.uri = uri

    def save(self) -> None:
        """Saves the config to the file.

        Returns
        -------
        None
            nothing
        """
        with fsspec.open(self.uri, mode="w", **self.auth) as f:
            self.write(f)

    def to_dict(self) -> Dict[str, Any]:
        """Converts config to a dictionary.

        Returns
        -------
        Dict[str, Any]
            Cofig in a dictionary format.
        """
        return {s: dict(self[s]) for s in self.keys() if self[s]}

    def read(self) -> "ExtendedConfigParser":
        """Reads config file.

        Returns
        -------
        ExtendedConfigParser
           Config parser object.
        """
        with fsspec.open(self.uri, "r", **self.auth) as f:
            self.read_string(f.read())
        return self

    def with_dict(self, info: Dict[str, Dict[str, Any]]) -> "ExtendedConfigParser":
        """Populates config with values from a dictionary.

        Parameters
        ----------
        info: Dict[str, Dict[str, Any]]
            Config in a dictionary format.

        Returns
        -------
        ExtendedConfigParser
           Config parser object.
        """
        self.read_dict(info)
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module created for the back compatability.
The original `model_metadata` was moved to the `ads.model` package.
"""

import warnings

warnings.warn(
    (
        "The `ads.common.model_metadata` is deprecated in `oracle-ads 2.6.8` "
        "and will be removed in future release. "
        "Use the `ads.model.model_metadata` instead."
    ),
    DeprecationWarning,
    stacklevel=2,
)

from ads.model.model_metadata import *
